/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package gluonj.deployment;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.jboss.deployment.DeploymentException;
import org.jboss.deployment.DeploymentInfo;
import org.jboss.deployment.SubDeployerSupport;
import org.jboss.logging.Logger;
import org.jboss.mx.loading.LoaderRepository;
import org.jboss.util.loading.Translator;

public class GlueDeployer extends SubDeployerSupport implements
        GlueDeployerMBean {

    private List deployedGlue;

    public GlueDeployer() {
        setSuffixes(new String[] { ".glue", "glue.xml" });
        deployedGlue = new LinkedList();
    }

    public boolean accepts(DeploymentInfo di) {

        String urlStr = di.url.toString();
        boolean b = urlStr.endsWith(".glue") || urlStr.endsWith("glue.xml");
        return b;
    }

    public static LoaderRepository r;

    public void create(DeploymentInfo di) throws DeploymentException {

        deployedGlue.add(di.localUrl);

        LoaderRepository rep = di.ucl.getLoaderRepository();
        r = rep;
        if (rep.getTranslator() == null) {
            rep.setTranslator(new GluonJTranslator(log));
        }
        Translator tr = rep.getTranslator();

        if (tr instanceof GluonJTranslator) {
            GluonJTranslator jbossTr = (GluonJTranslator) tr;

            List urlList = findGlueUrl(di, di.localCl);
            for (Iterator iter = urlList.iterator(); iter.hasNext();) {
                jbossTr.deployGlue((URL) iter.next());
            }
        }
        super.create(di);
    }

    private List findGlueUrl(DeploymentInfo di, URLClassLoader localCl)
            throws DeploymentException {

        try {
            List glueURLs = new LinkedList();
            if (di.isXML == true) {
                glueURLs.add(di.localUrl);
            } else {

                URL[] glueURL = new FileFinderInJar(di.localUrl)
                        .getFileURL("**/glue.xml");
                return Arrays.asList(glueURL);
            }
            return glueURLs;
        } catch (IOException e) {
            throw new DeploymentException(e);
        }
    }

    private URL getDocUrl(DeploymentInfo di, URLClassLoader localCL)
            throws DeploymentException {
        URL docURL = di.localUrl;
        if (di.isXML == false)
            docURL = localCL.findResource("META-INF/glue.xml");

        // Validate that the descriptor was found
        if (docURL == null)
            throw new DeploymentException("Failed to find META-INF/glue.xml");
        return docURL;
    }

    public void start(DeploymentInfo di) throws DeploymentException {
        log.debug("start.");

        super.start(di);
    }

    public void init(DeploymentInfo di) throws DeploymentException {
        log.debug("init.");

        try {
            if (di.watch == null) {
                // resolve the watch
                if (di.url.getProtocol().equals("file")) {
                    File file = new File(di.url.getFile());

                    // If not directory we watch the package
                    if (!file.isDirectory()) {
                        di.watch = di.url;
                    }
                    // If directory we watch the xml files
                    else {
                        di.watch = new URL(di.url, "META-INF/glue.xml");
                    }
                } else {
                    // We watch the top only, no directory support
                    di.watch = di.url;
                }
            }
        } catch (MalformedURLException e) {
            // TODO Auto-generated catch block
            throw new DeploymentException(e);
        }
        super.init(di);
    }

    public void destroy(DeploymentInfo di) throws DeploymentException {
        super.destroy(di);
    }

    public void stop(DeploymentInfo di) throws DeploymentException {

        log.info("stop. " + di);

        LoaderRepository rep = di.ucl.getLoaderRepository();
        Translator tr = rep.getTranslator();
        if (tr instanceof GluonJTranslator) {
            GluonJTranslator jbossTr = (GluonJTranslator) tr;
            jbossTr.undeployGlue(getDocUrl(di, di.localCl));

            deployedGlue.remove(di.localUrl);

            if (deployedGlue.size() == 0) {
                rep.setTranslator(null);
            }
        }
        super.stop(di);
    }

    protected ObjectName getObjectName(MBeanServer server, ObjectName name)
            throws MalformedObjectNameException {
        return name == null ? OBJECT_NAME : name;
    }

    protected void startService() throws Exception {
        log.info("Service Started.");
        super.startService();
    }

    protected void stopService() throws Exception {

        log.info("Service Stopped.");
        super.stopService();
    }
}