/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package gluonj.deployment;

import gluonj.ast.Aspect;
import gluonj.parser.AspectParser;
import gluonj.parser.ParseException;
import gluonj.weaver.AOPWeaver;
import gluonj.weaver.AOPWeaverFactory;
import gluonj.weaver.CannotCompileException;
import gluonj.weaver.InternalError;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.URL;
import java.security.CodeSource;
import java.security.ProtectionDomain;
import java.security.cert.Certificate;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.LoaderClassPath;

import org.jboss.logging.Logger;
import org.jboss.mx.loading.RepositoryClassLoader;
import org.jboss.util.loading.Translator;

import sun.misc.Resource;
import sun.misc.URLClassPath;

public class GluonJTranslator extends BugFixedTranslator {

    private Logger log;

    private AspectParser parser;

    private Map weaverMap;

    public GluonJTranslator(Logger l) {
        log = l;
        parser = AspectParser.factory();
        weaverMap = new TreeMap();
    }

    private AOPWeaver[] getDeployedGlueWeaver() {

        Set weaverEntrySet = weaverMap.entrySet();
        AOPWeaver[] weavers = new AOPWeaver[weaverEntrySet.size()];
        int i = 0;
        for (Iterator iter = weaverEntrySet.iterator(); iter.hasNext();) {
            weavers[i] = (AOPWeaver) ((Map.Entry) iter.next()).getValue();
            i++;
        }
        return weavers;
    }

    public byte[] transform0(ClassLoader cl, String name, Class clazz,
            ProtectionDomain domain, byte[] bytecode) throws Exception {

        try {
            log.debug("BEGIN transform(). name =" + name
                    + ", ProtectionDomain = " + domain);
//            setCorrectProtectionDomain(cl, name, clazz, domain);
            if(noWeaver()) {
                return bytecode;
            }

            if (isNonWeavedClassName(name)) {
                return null;
            }
            AOPWeaver[] weavers = getDeployedGlueWeaver();
            ClassPool pool = getCachedPool(cl, weavers);
            CtClass cc = pool.get(name);
            for (int i = 0; i < weavers.length; i++) {
                AOPWeaver weaver = weavers[i];

                Set s = new TreeSet();
                weaver.setCreatedClassSet(s);
                
                weaver.weaveIntertypeDecl(cc, pool);
                weaver.weaveAdvice(cc, pool);

                for (Iterator iter = s.iterator(); iter.hasNext();) {
                    defineClass(cl, (CtClass) iter.next());
                }
            }

            return cc.toBytecode();
        } catch (InternalError e) {
            log.error("Failed to weave to " + name + ". cause: " + e);
            throw e;
        } catch (Throwable t) {
            log.debug("Exception was thrown: " + t);
            t.printStackTrace();
            throw new Exception(t);
        }
    }

    private boolean noWeaver() {
        return weaverMap.entrySet().size() == 0;
    }

    /*
     * HashMap <ClassLoader, ClassPool>
     */
    private Map cpTable = new HashMap();

    private ClassPool getCachedPool(ClassLoader cl, AOPWeaver[] w) {
        
        if(!cpTable.containsKey(cl)) {
            ClassPool pool = new ClassPool(ClassPool.getDefault());
            pool.appendClassPath(new LoaderClassPath(cl));
            cpTable.put(cl, pool);
        }
        return (ClassPool) cpTable.get(cl);
    }

    private void defineClass(ClassLoader cl, CtClass cc) throws Exception {

        Method mDefineClass = ClassLoader.class.getDeclaredMethod(
                "defineClass", new Class[] { String.class,
                        new byte[0].getClass(), Integer.TYPE, Integer.TYPE });

        mDefineClass.setAccessible(true);
        byte[] bytecode = cc.toBytecode();
        mDefineClass.invoke(cl, new Object[] { cc.getName(), bytecode,
                new Integer(0), new Integer(bytecode.length) });
    }

    private void setCorrectProtectionDomain(ClassLoader cl, String name,
            Class clazz, ProtectionDomain domain) throws Exception {

        RepositoryClassLoader rcl = ((RepositoryClassLoader) cl);
        String s = name.replace('.', '/').concat(".class");

        /*
         * BEGIN URLClassPath ucp = cl.ucp;
         */
        Class rclc = Class
                .forName("org.jboss.mx.loading.RepositoryClassLoader");
        Class uclc = Class.forName("java.net.URLClassLoader");
        Field fUcp = uclc.getDeclaredField("ucp");
        fUcp.setAccessible(true);
        URLClassPath ucp = (URLClassPath) fUcp.get(rcl);
        /*
         * END
         */
        Resource resource = ucp.getResource(s, false);
        URL url = resource.getCodeSourceURL();
        resource.getBytes(); // this method creates certificates.
        Certificate[] acertificate = resource.getCertificates();
        CodeSource codesource = new CodeSource(url, acertificate);
        /*
         * BEGIN ProtectionDomain newdomain =
         * rcl.getProtectionDomain(codesource);
         */
        Class cSecureClassLoader = Class
                .forName("java.security.SecureClassLoader");

        Method[] ms = cSecureClassLoader.getDeclaredMethods();
        Method mGetProtectionDomain = cSecureClassLoader.getDeclaredMethod(
                "getProtectionDomain", new Class[] { Class
                        .forName("java.security.CodeSource"), });

        mGetProtectionDomain.setAccessible(true);
        ProtectionDomain newdomain = (ProtectionDomain) mGetProtectionDomain
                .invoke(rcl, new Object[] { codesource });
        /*
         * END
         */
        /*
         * BEGIN domain.codesource = newdomain.codesource;
         */
        Class cProtectionDomain = Class
                .forName("java.security.ProtectionDomain");
        Field fCodesource = cProtectionDomain.getDeclaredField("codesource");
        fCodesource.setAccessible(true);
        fCodesource.set(domain, fCodesource.get(newdomain));
        /*
         * END
         */
    }

    private boolean isNonWeavedClassName(String classname) {
        return (classname.startsWith("gluonj")
                || classname.startsWith("javassist")
                || classname.startsWith("org.jboss")
                || classname.startsWith("org.jgroups")
                || classname.startsWith("org.apache.crimson")
                || classname.startsWith("org.apache.xalan")
                || classname.startsWith("org.apache.xml")
                || classname.startsWith("org.apache.xpath")
                || classname.startsWith("org.ietf.")
                || classname.startsWith("org.omg.")
                || classname.startsWith("org.w3c.")
                || classname.startsWith("org.xml.sax.")
                || classname.startsWith("sunw.")
                || classname.startsWith("sun.")
                || classname.startsWith("java.")
                || classname.startsWith("javax.")
                || classname.startsWith("com.sun.") || classname
                .startsWith("junit"));
    }

    public void unregisterClassLoader(ClassLoader arg0) {
    }

    public void deployGlue(URL glueUrl) {

        log.info("GluonJTranslator.deploy(), URL=" + glueUrl);

        try {
            Aspect a = parser.parse(glueUrl);
            log.debug(a);
            AOPWeaver weaver = new AOPWeaverFactory().make(new Aspect[] { a });
            weaverMap.put(glueUrl.toString(), weaver);
        } catch (ParseException e) {
            log.error(e);
        } catch (CannotCompileException e) {
            log.error(e);
        }
    }

    public void undeployGlue(URL glueUrl) {
        log.info("GluonJTranslator.undeploy(), URL=" + glueUrl);
        weaverMap.remove(glueUrl.toString());
    }
}