/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.ast;

import gluonj.ast.java.ASTBlock;
import gluonj.ast.java.ASTFormalParameter;
import gluonj.ast.joinpoint.JoinPoint;
import gluonj.ast.pcs.PointcutResidue;
import gluonj.ast.pointcut.ASTPCShadow;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.ast.pointcut.visitor.DNFVisitor;
import gluonj.ast.pointcut.visitor.Traverser;

import java.util.LinkedList;
import java.util.List;

class AdviceInfoImpl implements AdviceInfo {

    private String body;

    private AdviceType type;

    private AdviceInfoImpl(String s, AdviceType t) {
        body = s;
        type = t;
    }

    public String getBody() {
        return body;
    }

    public AdviceType getType() {
        return type;
    }

    public static AdviceInfo factory(Advice advice, JoinPoint jp) {
        AdviceInfoImpl info = new Factory().make(advice, jp);
        return info;
    }

    private static class Factory {

        private boolean[] matchEachConjunction;

        private ASTPCShadow[][] init(Advice advice) {
            ASTPointcutExpr root = advice.getPointcutExpr();
            CompiledPCGetter g = new CompiledPCGetter();
            Traverser.traverseDNF(root, g);
            ASTPCShadow[][] pointcut = g.getPointcut();
            matchEachConjunction = new boolean[pointcut.length];
            return pointcut;
        }

        AdviceInfoImpl make(Advice advice, JoinPoint jp) {

            ASTPCShadow[][] pointcut = init(advice);

            if (match(pointcut, jp)) {
                String body = makeBody(advice.getBody(), advice
                        .getAllParameter(), pointcut, jp);
                return new AdviceInfoImpl(body, advice.getType());
            } else {
                return null;
            }
        }

        private boolean match(ASTPCShadow[][] pointcut, JoinPoint jp) {
            boolean match = false;
            for (int i = 0; i < pointcut.length; i++) {
                matchEachConjunction[i] = match(pointcut[i], jp);
                match = match || matchEachConjunction[i];
            }
            return match;
        }

        private boolean match(ASTPCShadow[] conjunction, JoinPoint jp) {
            for (int i = 0; i < conjunction.length; i++) {
                PointcutResidue r = conjunction[i].match(jp);
                if (r == PointcutResidue.FALSE) {
                    return false;
                }
            }
            return true;
        }

        private String makeBody(ASTBlock body, ASTFormalParameter[] parameter,
                ASTPCShadow[][] pointcut, JoinPoint jp) {

            StringBuffer buf = new StringBuffer();

            for (int i = 0; i < parameter.length; i++) {
                String type = parameter[i].getType().getType();
                String name = parameter[i].getName();
                if (isPrimitive(type)) {
                    buf.append(type + " " + name + ";\n");
                } else {
                    buf.append(type + " " + name + "= null;\n");
                }
            }

            PointcutResidue[][] residues = matchPointcut(pointcut, jp);

            boolean alwaysTrue = isAlwaysTrue(residues, jp);

            if (alwaysTrue) {
                buf.append(makeAlwaysTrueBody(residues, body));
            } else {
                buf.append(makeConditionalBody(residues, jp, body));
            }

            return "{" + buf.toString() + "\n}";
        }

        private String makeConditionalBody(PointcutResidue[][] residues,
                JoinPoint jp, ASTBlock body) {
            StringBuffer buf = new StringBuffer();

            buf.append("boolean weaveOK = false;\n");

            boolean addElse = false;
            for (int i = 0; i < residues.length; i++) {
                if (matchEachConjunction[i]) {

                    StringBuffer condition = new StringBuffer("true ");
                    StringBuffer statement = new StringBuffer();

                    for (int j = 0; j < residues[i].length; j++) {
                        PointcutResidue r = residues[i][j];
                        condition.append(" &&" + r.getDynamicCondition(jp));
                        statement.append(r.getStatement(jp));
                    }

                    if (addElse) {
                        buf.append("else ");
                    }
                    buf.append("if(" + condition + ") {\n");
                    buf.append("{" + statement + "\n}\n");
                    buf.append("weaveOK = true;\n");
                    buf.append("}\n");
                    addElse = true;
                }
            }

            buf.append("if(weaveOK) {\n");
            buf.append(body.toString());
            buf.append("\n}");

            return "{" + buf.toString() + "\n}";
        }

        private String makeAlwaysTrueBody(PointcutResidue[][] residues,
                ASTBlock body) {
            return body.toString();
        }

        private PointcutResidue[][] matchPointcut(ASTPCShadow[][] pointcut,
                JoinPoint jp) {
            PointcutResidue[][] residues = new PointcutResidue[pointcut.length][];
            for (int i = 0; i < pointcut.length; i++) {
                residues[i] = new PointcutResidue[pointcut[i].length];
                for (int j = 0; j < pointcut[i].length; j++) {
                    residues[i][j] = pointcut[i][j].match(jp);
                }
            }
            return residues;
        }

        private boolean isAlwaysTrue(PointcutResidue[][] residues, JoinPoint jp) {
            for (int i = 0; i < residues.length; i++) {
                if (matchEachConjunction[i]) {
                    boolean b = true;
                    for (int j = 0; j < residues[i].length; j++) {
                        PointcutResidue r = residues[i][j];
                        if (!r.getDynamicCondition(jp).trim().equals("true")) {
                            b = false;
                            break;
                        }
                    }
                    if (b) {
                        return true;
                    }
                }
            }
            return false;
        }

        private boolean isPrimitive(String type) {
            return type.equals("boolean") || type.equals("byte")
                    || type.equals("char") || type.equals("double")
                    || type.equals("float") || type.equals("int")
                    || type.equals("long") || type.equals("short")
                    || type.equals("void");
        }

        private class CompiledPCGetter extends DNFVisitor {

            private List pcList = new LinkedList();

            public void atConjunctive(List exprs) {
                ASTPCShadow[] shadows = (ASTPCShadow[]) exprs
                        .toArray(new ASTPCShadow[exprs.size()]);
                pcList.add(shadows);
            }

            public ASTPCShadow[][] getPointcut() {
                return (ASTPCShadow[][]) pcList.toArray(new ASTPCShadow[pcList
                        .size()][]);
            }
        }
    }

}