/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.ast.pattern;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

public class SubTypeNamePattern extends BasicTypeNamePattern {

    private BasicTypeNamePattern child;

    public SubTypeNamePattern(BasicTypeNamePattern c) {
        child = c;
    }

    public String toString() {
        return child + "+";
    }

    public BasicTypeNamePattern getChild() {
        return child;
    }

    public void accept(TypeNamePatternVisitor v) {
        v.atSubTypeName(this);
    }

    public boolean match(String typename) {
        String[] parentNames = getParentTypeNames(typename);

        if (match(parentNames))
            return true;

        String[] interfaceNames = getInterfaceNames(typename);

        if (match(interfaceNames))
            return true;

        return false;
    }

    private boolean match(String[] typeNames) {
        for (int i = 0; i < typeNames.length; i++) {
            String name = typeNames[i];
            boolean b = child.match(name);
            if (b) {
                return true;
            }
        }
        return false;

    }

    private String[] getParentTypeNames(String name) {

        /*
         * Array
         */
        if (name.endsWith("[]")) {
            return new String[] { "java.lang.Object", name };
        }

        /*
         * Primitive
         */
        if (isPrimitive(name)) {
            return new String[] { name };
        }

        /*
         * Class
         */
        try {
            List parents = new LinkedList();
            Class currentParent = Class.forName(name);
            do {
                parents.add(currentParent.getName());
                currentParent = currentParent.getSuperclass();
            } while (currentParent != null);
            return (String[]) parents.toArray(new String[parents.size()]);

        } catch (ClassNotFoundException e) {
            return new String[0];
        }
    }

    private String[] getInterfaceNames(String name) {
        if (!name.endsWith("[]") && !isPrimitive(name)) {
            /*
             * Class
             */
            try {
                Set interfaceNames = new TreeSet();

                String[] parentNames = getParentTypeNames(name);

                for (int i = 0; i < parentNames.length; i++) {
                    String parentName = parentNames[i];
                    Class[] interfaces = Class.forName(parentName)
                            .getInterfaces();
                    for (int j = 0; j < interfaces.length; j++) {
                        interfaceNames.add(interfaces[j].getName());
                    }
                }
                return (String[]) interfaceNames
                        .toArray(new String[interfaceNames.size()]);

            } catch (ClassNotFoundException e) {
                return new String[0];
            }
        } else {
            return new String[0];
        }
    }

    private boolean isPrimitive(String type) {
        return type.equals("boolean") || type.equals("byte")
                || type.equals("char") || type.equals("double")
                || type.equals("float") || type.equals("int")
                || type.equals("long") || type.equals("short")
                || type.equals("void");
    }

}