/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.parser;

import gluonj.ast.pointcut.ASTAnd;
import gluonj.ast.pointcut.ASTNot;
import gluonj.ast.pointcut.ASTOr;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.ast.pointcut.ASTPointcutPrimitive;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public abstract class AbstractPointcutParser implements TokenId {

    protected Lexer lexer;

    private PcdFactoryTable pcdFactoryTable = initTable();

    public AbstractPointcutParser() {
        super();
    }

    public AbstractPointcutParser(Lexer lex) {
        lexer = lex;
    }

    public abstract PcdFactoryTable initTable();

    /*
     * Tokens used in this class : ( ) || && ! , . * + IDENTIFIER THROWS
     */

    public ASTPointcutExpr parsePointcut() throws ParseException {

        return parsePointcut(new Context());
    }

    /*
     * pointcut := pointcut.or
     */

    public ASTPointcutExpr parsePointcut(Context c) throws ParseException {

        return parsePointcutOr(c);
    }

    /*
     * pointcut.or := pointcut.and (("||" | "OR" ) pointcut.and)*
     */

    private ASTPointcutExpr parsePointcutOr(Context c) throws ParseException {

        ASTPointcutExpr e = parsePointcutAnd(c);

        while (lexer.lookAhead() == OROR
                || (lexer.lookAhead() == Identifier && lexer.getIdentifier()
                        .equals("OR"))) {

            lexer.get();
            e = new ASTOr(e, parsePointcutAnd(c));
        }
        return e;
    }

    /*
     * pointcut.and := pointcut.not (("&&" | "AND" ) pointcut.not)*
     */

    private ASTPointcutExpr parsePointcutAnd(Context c) throws ParseException {

        ASTPointcutExpr e = parsePointcutNot(c);

        while (lexer.lookAhead() == ANDAND
                || (lexer.lookAhead() == Identifier && lexer.getIdentifier()
                        .equals("AND"))) {
            lexer.get();
            e = new ASTAnd(e, parsePointcutNot(c));
        }
        return e;
    }

    /*
     * pointcut.not := ["!"]? pointcut.unary
     */
    private ASTPointcutExpr parsePointcutNot(Context c) throws ParseException {

        if (lexer.lookAhead() == '!') {
            lexer.get();
            return new ASTNot(parsePointcutUnary(c));
        } else {
            return parsePointcutUnary(c);
        }
    }

    /*
     * pointcut.unary := "(" pointcut ")" | pointcut.designater
     */
    private ASTPointcutExpr parsePointcutUnary(Context c) throws ParseException {

        if (lexer.lookAhead() == '(') {
            lexer.get();
            ASTPointcutExpr expr = parsePointcut(c);
            if (lexer.lookAhead() == ')') {
                lexer.get();
                return expr;
            } else
                throw parseError(c, new int[] { ')' });
        } else {
            return parsePointcutDesignator(c);
        }
    }

    /*
     * pointcut.designator := (pointcut.identifier | "this") "("
     * pointcut.arguments ")"
     */

    private ASTPointcutExpr parsePointcutDesignator(Context c)
            throws ParseException {

        String pcdName;
        pcdName = parsePointcutIdentifier(c);
        PcdFactory factory = pcdFactoryTable.getPcdFactory(pcdName);

        if (factory == null) {
            throw parseError("The name of pointcut designator is wrong: "
                    + pcdName);
        }

        ASTPointcutExpr expr = factory.makePcd(c);
        return expr;
    }

    private String parsePointcutIdentifier(Context c) throws ParseException {
        if (lexer.lookAhead() == Identifier) {
            String id = lexer.getIdentifier();
            lexer.get();
            return id;
        }
        throw parseError(c, new int[] { Identifier });
    }

    private ParseException parseError(Context c, int[] expects) {
        return parseError(" Syntax of pointcut is wrong. " + idToString(expects)
                + " are expected.");
    }

    
    private ParseException parseError(String message) {
        return new ParseException(message + "\n" + "Lexer dump : "
                + lexer.toString());
    }

    private String idToString(int tokenId) {
        if (tokenId == Identifier) {
            return "\"Identifier\"";
        } else {
            return "\"" + ((char) tokenId) + "\"";
        }
    }

    private String idToString(int[] tokenIds) {
        StringBuffer buf = new StringBuffer();

        if (tokenIds.length > 0) {
            buf.append(idToString(0));
        }

        for (int i = 1; i < tokenIds.length; i++) {
            buf.append(",");
            buf.append(idToString(i));
        }
        return buf.toString();
    }

    public static class Context {
    }

    public static class PcdFactoryTable {

        private Map m = new HashMap();

        public void putPcdFactory(PcdFactory spec) {
            m.put(spec.getId(), spec);
        }

        public PcdFactory getPcdFactory(String pcdname) {
            return ((PcdAbstractFactory) m.get(pcdname));
        }
    }

    public static interface PcdFactory {
        public String getId();

        public ASTPointcutPrimitive makePcd(Context c) throws ParseException;
    }

    public abstract class PcdAbstractFactory implements PcdFactory {

        public abstract String getId();

        private Object[] parsePointcutArguments(Context c)
                throws ParseException {

            if (lexer.lookAhead() != '(') {
                throw parseError("Syntax error. ( expected.");
            } else {
                lexer.get();
            }
            if (lexer.lookAhead() == ')') {
                lexer.get();
                return new Object[0];
            }

            List args = new LinkedList();
            args.add(parsePointcutArgument(c));
            while (lexer.lookAhead() == ',') {
                lexer.get();
                args.add(parsePointcutArgument(c));
            }
            if (lexer.lookAhead() != ')') {
                throw parseError("Syntax error. ) expected.");
            } else {
                lexer.get();
                return args.toArray(new Object[args.size()]);
            }
        }

        public ASTPointcutPrimitive makePcd(Context c) throws ParseException {
            return make(parsePointcutArguments(c), c);
        }

        public abstract ASTPointcutPrimitive make(Object[] args, Context c)
                throws ParseException;

        public abstract Object parsePointcutArgument(Context c)
                throws ParseException;

        public ParseException getArgsNumException(int validNum)
                throws ParseException {
            throw parseError("The number of the arguments of " + getId()
                    + " pointcut must be " + validNum + ".");
        }
    }

}