/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.parser;

import gluonj.ast.Advice;
import gluonj.ast.AdviceType;
import gluonj.ast.Aspect;
import gluonj.ast.Import;
import gluonj.ast.PointcutDecl;
import gluonj.ast.intertype.IntertypeDecl;
import gluonj.ast.java.ASTBlock;
import gluonj.ast.java.ASTFormalParameter;
import gluonj.ast.java.ASTType;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.ast.xml.ElmAdvice;
import gluonj.ast.xml.ElmAdviceAfter;
import gluonj.ast.xml.ElmAdviceAround;
import gluonj.ast.xml.ElmAdviceBefore;
import gluonj.ast.xml.ElmAspect;
import gluonj.ast.xml.ElmImport;
import gluonj.ast.xml.ElmInjection;
import gluonj.ast.xml.ElmParam;
import gluonj.ast.xml.ElmPointcutDecl;
import gluonj.ast.xml.RVisitorBase;
import gluonj.ast.xml.URVisitor;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

public class AspectParser {

    public static AspectParser factory() {
        if (singleton == null) {
            singleton = new AspectParser();
        }
        return singleton;
    }

    private static AspectParser singleton;

    protected AspectParser() {
        aspectMap = new HashMap();
    }

    private HashMap aspectMap;

    public Aspect parse(URL url) throws ParseException {
        ElmAspect a;
        try {
            a = new ElmAspect(url);
        } catch (SAXException e) {
            throw new ParseException(e);
        } catch (ParserConfigurationException e) {
            throw new ParseException(e);
        } catch (IOException e) {
            throw new ParseException(e);
        } catch (RuntimeException e) {
            if (e.getCause() instanceof ParseException) {
                throw (ParseException) e.getCause();
            } else {
                throw e;
            }
        }
        return parseInner(a);
    }

    public Aspect parse(String filePath) throws ParseException {
        return parseInner(filePath);
    }

    protected Aspect parseInner(String filePath) throws ParseException {

        ElmAspect a;
        try {
            a = new ElmAspect(new File(filePath));
        } catch (SAXException e) {
            throw new ParseException(e);
        } catch (ParserConfigurationException e) {
            throw new ParseException(e);
        } catch (IOException e) {
            throw new ParseException(e);
        } catch (RuntimeException e) {
            if (e.getCause() instanceof ParseException) {
                throw (ParseException) e.getCause();
            } else {
                throw e;
            }
        }
        return parseInner(a);
    }

    protected Aspect parseInner(ElmAspect a) throws ParseException {

        AstBuilder builder = new AstBuilder();
        URVisitor.traverse(a, builder);
        Aspect aspect = builder.build();
        return aspect;
    }

    private class AstBuilder extends RVisitorBase {

        public Aspect build() {
            return aspect;
        }

        private Aspect aspect;

        private List importList;

        private List injectionDeclList;

        private List pointcutDeclList;

        private List adviceList;

        private List parameterList;

        private AdviceType adviceType;

        private ASTBlock adviceBody;

        public boolean enter(ElmAspect visitable) {
            injectionDeclList = new LinkedList();
            pointcutDeclList = new LinkedList();
            adviceList = new LinkedList();
            importList = new LinkedList();
            return true;
        }

        public void leave(ElmAspect visitable) {
            aspect = new Aspect(importList, injectionDeclList,
                    pointcutDeclList, adviceList);
        }

        public boolean enter(ElmAdvice advice) {
            parameterList = new LinkedList();
            adviceType = null;
            adviceBody = null;
            return true;
        }

        public void leave(ElmAdvice advice) {
            ASTFormalParameter[] param = (ASTFormalParameter[]) parameterList
                    .toArray(new ASTFormalParameter[parameterList.size()]);
            ASTPointcutExpr root = parse(advice.getPointcut().trim());
            adviceList.add(new Advice(param, adviceType, root, adviceBody));
        }

        ASTPointcutExpr parse(String expr) {
            PointcutParser.Context c = new PointcutParser.Context();
            for (Iterator iter = parameterList.iterator(); iter.hasNext();) {
                c.addPointcutParam((ASTFormalParameter) iter.next());
            }
            String[] pckName = new String[importList.size()];
            int i=0;
            for (Iterator iter = importList.iterator(); iter.hasNext();i++) {
                pckName[i] = ((Import)iter.next()).getPath();
            }
            c.setImportedPackage(pckName);
            try {
                ASTPointcutExpr e = new PointcutParser(new Lexer(expr)).parsePointcut(c);
                c.clearImportedPackage();
                return e;
            } catch (ParseException e) {
                throw new RuntimeException(e);
            }
        }

        public boolean enter(ElmParam param) {
            parameterList.add(new ASTFormalParameter(new ASTType(param
                    .getType().trim()), param.getName().trim()));
            return false;
        }

        public boolean enter(ElmAdviceBefore adv) {
            adviceBody = new ASTBlock(adv.getContent().trim());
            adviceType = AdviceType.BEFORE;
            return false;
        }

        public boolean enter(ElmAdviceAround adv) {
            adviceBody = new ASTBlock(adv.getContent().trim());
            adviceType = AdviceType.AROUND;
            return false;
        }

        public boolean enter(ElmAdviceAfter adv) {
            adviceBody = new ASTBlock(adv.getContent().trim());
            adviceType = AdviceType.AFTER;
            return false;
        }

        public boolean enter(ElmInjection injection) {

            try {
                String decls = injection.getContent().trim();
                InjectionCodeParser parser = new InjectionCodeParser(new Lexer(
                        decls));
                IntertypeDecl[] decl = parser.parse();
                injectionDeclList.addAll(Arrays.asList(decl));
                return false;

            } catch (ParseException e) {
                throw new RuntimeException(e);
            }
        }

        public boolean enter(ElmPointcutDecl visitable) {
            parameterList = new LinkedList();
            return true;
        }

        public void leave(ElmPointcutDecl decl) {
            pointcutDeclList.add(new PointcutDecl(decl.getName().trim(),
                    parameterList, parse(decl.getPointcut().trim())));
        }

        public boolean enter(ElmImport visitable) {
            String pckName = visitable.getContent().trim();
            if (pckName.equals("") || pckName.matches(".*\\.[\\.]+.*")
                    || pckName.startsWith(".") || pckName.endsWith(".")) {
                throw new RuntimeException(new ParseException(
                        "package name is wrong: " + pckName));
            }

            String[] n = pckName.split("\\.");
            for (int i = 0; i < n.length; i++) {
                if (!isJavaName(n[i])) {
                    throw new RuntimeException(new ParseException(
                            "wrong java identifier: " + n[i]));
                }
            }
            importList.add(new Import(pckName));
            return true;
        }

        private boolean isJavaName(String name) {

            if (name.length() == 0) {
                return false;
            }
            if (!Character.isJavaIdentifierStart(name.charAt(0))) {
                return false;
            }
            for (int i = 1; i < name.length(); i++) {
                if (!Character.isJavaIdentifierPart(name.charAt(i))) {
                    return false;
                }
            }
            return true;
        }
    }
}