/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.parser;

import gluonj.ast.intertype.IntertypeDecl;
import gluonj.ast.java.ASTBehavior;
import gluonj.ast.java.ASTBlock;
import gluonj.ast.java.ASTConstructor;
import gluonj.ast.java.ASTExpr;
import gluonj.ast.java.ASTField;
import gluonj.ast.java.ASTFormalParameter;
import gluonj.ast.java.ASTMember;
import gluonj.ast.java.ASTMethod;
import gluonj.ast.java.ASTPrimitiveType;
import gluonj.ast.java.ASTType;
import gluonj.ast.java.Modifier;
import gluonj.ast.java.ext.ASTAnonymousField;
import gluonj.ast.java.ext.ASTCflowType;
import gluonj.ast.pattern.DotTypeNamePattern;
import gluonj.ast.pattern.IdPattern;
import gluonj.ast.pattern.IdentifierTypeNamePattern;
import gluonj.ast.pattern.PrimitiveTypeNamePattern;
import gluonj.ast.pattern.SingleTypeNamePattern;
import gluonj.ast.pattern.TypeNamePattern;
import gluonj.ast.pcs.PcsModifier;
import gluonj.ast.pcs.PcsType;
import gluonj.ast.pointcut.ASTPointcutExpr;

import java.util.LinkedList;
import java.util.List;

import javassist.compiler.CompileError;
import javassist.compiler.Parser;
import javassist.compiler.SymbolTable;
import javassist.compiler.TokenId;
import javassist.compiler.ast.ASTree;
import javassist.compiler.ast.Stmnt;

public class InjectionCodeParser implements TokenId {

    private static final String ANONYMOUS_FIELD_NAME = "aspect";

    private static final String CFLOW = "Cflow";

    String decls;

    Lexer lexer;

    public InjectionCodeParser(Lexer l) {
        lexer = l;
        //		logLexer();
    }

    private void logLexer() {
        int i = 0;
        while (i != -1) {
            System.out.println((int) lexer.lookAhead());
            System.out.println((char) lexer.lookAhead());
            System.out.println(lexer.getIdentifier());
            i = lexer.get();
            System.out.println(i);
            System.out.println("##############");
        }
    }

    /*
     * List <InjectionDeclaration>
     */
    public IntertypeDecl[] parse() throws ParseException {
        try {
            return parseInjectionDeclarations(new Context());
        } catch (ParseException e) {
            System.out.println(e);
            throw e;
        }
    }

    /*
     * injection.decls := injection.decl*
     */
    private IntertypeDecl[] parseInjectionDeclarations(Context c)
            throws ParseException {
        List declList = new LinkedList();
        while (lexer.lookAhead() != -1) {
            IntertypeDecl decl = parseInjectionDeclaration(c);
            declList.add(decl);
        }
        return (IntertypeDecl[]) declList.toArray(new IntertypeDecl[declList
                .size()]);
    }

    /*
     * constructor: java.modifiers java.type ".new(" java.type* ")" ("throws"
     * java.type+) "{" body "}"
     * 
     * method: java.modifiers java.type java.type "." name "(" java.type* ")"
     * ("throws" java.type+) "{" body "}"
     * 
     * field: java.modifiers java.type java.type "." name "=" initializer
     * 
     * field:(an,cf)java.modifiers java.type "Cflow(" pc ")" "." "aspect" "="
     * initializer
     * 
     * field:(an,as)java.modifiers java.type java.type "." "aspect" "("
     * java.type* ")" "=" initializer
     */

    /*
     * injection.decl := java.modifiers java.type injected.type&aspect
     * (parameter)? ( "=" initializer)? ";"
     */
    private IntertypeDecl parseInjectionDeclaration(Context c)
            throws ParseException {

        Modifier mod = parseJavaModifiers(c);
        ASTType returnType = parseJavaTypeBeforeLastId(c);
        ASTType targetType = null;
        String name = null;
        ASTFormalParameter[] param = null;
        ASTType[] throwable = null;
        ASTExpr initializer = null;
        String body = null;

        /* pattern is constructor */
        if ((lexer.lookAhead() == '.')) {
            targetType = returnType;
            returnType = null;
        } else {
            targetType = parseDeclaringClassType(c);
        }
        /* parse member name */
        if (lexer.lookAhead() == '.') {
            name = parseMemberName(c);
        }
        /* read parameter */
        if (lexer.lookAhead() == '(') {
            param = parseParameters(c);
        }
        /* read exception */
        if (lexer.lookAhead() == THROWS) {
            throwable = parseThrowable(c);
        }
        /* read field initializer or constructor/method body */
        if (lexer.lookAhead() == '{') {
            body = parseBehaviorBody();
        } else {
            if (lexer.lookAhead() == '=') {
                lexer.get();//=
                initializer = parseInitializer();
            }
            if (lexer.lookAhead() == ';') {
                lexer.get();//;
            } else {
                throw parseError(c, ";");
            }
        }
        if (param != null && param.length > 1) {
            if (body != null && !allNonNullName(param)) {
                throw new ParseException(
                        "Parameter names are missing at method/constructor declaration.");
            }
            if (body == null && !allNullName(param)) {
                throw new ParseException(
                        "Writing parameter names at anonymous field declaration is not permitted.");
            }
        }

        IntertypeDecl decl;
        if (body == null && name.equals("aspect")) {
            decl = makeAnonymousFieldDecl(mod, returnType, param, throwable,
                    initializer, targetType);
        } else {
            PcsType target = makePcsType(targetType);
            ASTMember member;
            if (body != null) {
                member = makeBehavior(mod, returnType, name, param, throwable,
                        body);
            } else {
                member = makeField(mod, returnType, name, param, throwable,
                        initializer);
            }
            decl = new IntertypeDecl(target, member);
        }
        while (lexer.lookAhead() == ';') {
            lexer.get();//;
        }
        return decl;
    }

    private ASTExpr parseInitializer() throws ParseException {
        String initializer;
        try {
            lexer.startMark();
            ASTree t = new Parser(lexer.lex).parseExpression(new SymbolTable());
            initializer = lexer.endMark();
            return new ASTExpr(initializer);
        } catch (CompileError e) {
            throw new ParseException("fail to compile the field initializer.",
                    e);
        }
    }

    private String parseBehaviorBody() throws ParseException {
        try {
            String body = parseJavaStatement(lexer);
            return body;
        } catch (CompileError e) {
            throw new ParseException(
                    "fail to compile the method/constructor body.", e);
        }
    }

    private ASTType[] parseThrowable(Context c) throws ParseException {

        lexer.get();//THROWS
        return parseTypes(c);

    }

    private boolean allNullName(ASTFormalParameter[] param) {
        for (int i = 0; i < param.length; i++) {
            if (param[i].getName() != null)
                return false;
        }
        return true;
    }

    private boolean allNonNullName(ASTFormalParameter[] param) {
        for (int i = 0; i < param.length; i++) {
            if (param[i].getName() == null)
                return false;
        }
        return true;
    }

    private ASTMember makeField(Modifier mod, ASTType returnType, String name,
            ASTFormalParameter[] param, ASTType[] throwable, ASTExpr initializer)
            throws ParseException {

        checkFieldSig(returnType, name, throwable);
        if (param != null) {
            throw new ParseException(
                    "Syntax error. parameters are appeared at field declaration.");
        }
        return new ASTField(mod, returnType, name, initializer);
    }

    private void checkFieldSig(ASTType returnType, String name,
            ASTType[] throwable) throws ParseException {

        if (returnType == null) {
            throw new ParseException(
                    "return type is missing at field declaration.");
        }
        if (name.equals("new")) {
            throw new ParseException("new is wrong name.");
        }
        if (throwable != null) {
            throw new ParseException("\"throws\" is appeared at wrong place.");
        }
    }

    private IntertypeDecl makeAnonymousFieldDecl(Modifier mod,
            ASTType returnType, ASTFormalParameter[] param,
            ASTType[] throwable, ASTExpr initializer, ASTType target)
            throws ParseException {

        checkFieldSig(returnType, ANONYMOUS_FIELD_NAME, throwable);
        if (target instanceof ASTCflowType) {
            ASTAnonymousField af = new ASTAnonymousField(mod, returnType,
                    initializer, new ASTType[] { target });
            PcsType type = new PcsType(new PcsModifier(),
                    new IdentifierTypeNamePattern(new IdPattern("<cflow>")));
            return new IntertypeDecl(type, af);
        } else {
            ASTType[] types;
            if (param != null) {
                types = new ASTType[param.length];
                for (int i = 0; i < types.length; i++) {
                    types[i] = param[i].getType();
                }
            } else {
                types = new ASTType[0];
            }
            ASTAnonymousField af = new ASTAnonymousField(mod, returnType,
                    initializer, types);
            return new IntertypeDecl(makePcsType(target), af);
        }
    }

    private ASTMember makeBehavior(Modifier mod, ASTType returnType,
            String name, ASTFormalParameter[] param, ASTType[] throwable,
            String body) throws ParseException {
        ASTBehavior member;
        if (throwable == null) {
            throwable = new ASTType[0];
        }
        if (param == null) {
            param = new ASTFormalParameter[0];
        }
        if (returnType == null && name.equals("new")) {
            member = new ASTConstructor(mod, param, throwable, new ASTBlock(
                    body));
        } else if (returnType != null && !name.equals("new")) {
            member = new ASTMethod(mod, returnType, name, param, throwable,
                    new ASTBlock(body));
        } else {
            throw new ParseException("Syntax error. return type is missing. ");
        }
        return member;
    }

    private String parseMemberName(Context c) throws ParseException {
        String name;
        lexer.get();//.
        if (lexer.lookAhead() == Identifier) {
            name = lexer.getIdentifier();
            lexer.get();
        } else if (lexer.lookAhead() == NEW) {
            name = "new";
            lexer.get();
        } else {
            throw parseError(c, "identifier or new");
        }
        return name;
    }

    private IntertypeDecl parseInjectionDeclaration2(Context c)
            throws ParseException {

        Modifier mod = parseJavaModifiers(c);
        ASTType returnType = parseJavaTypeBeforeLastId(c);
        ASTType targetType = null;
        String name = null;

        /* !constructor pat */
        if (!(lexer.lookAhead() == '.')) {
            targetType = parseJavaTypeBeforeLastId(c);
        }
        if (lexer.lookAhead() == '.') {
            lexer.get();//.
            if (lexer.lookAhead() == Identifier) {
                name = lexer.getIdentifier();
                lexer.get();
            } else if (lexer.lookAhead() == NEW) {
                name = "new";
                lexer.get();
            }
        }

        /* read constructor or method declaration */
        if (lexer.lookAhead() == '(') {
            lexer.get();

            ASTFormalParameter[] param;
            ASTType[] throwable;
            String body = null;

            param = parseParameters(c);
            if (lexer.lookAhead() == ')') {
                lexer.get();//')'
            } else
                throw new ParseException("\')\' is not found." + lexer);

            if (lexer.lookAhead() == THROWS) {
                lexer.get();// "throws"
                throwable = parseTypes(c);
            } else {
                throwable = new ASTType[0];
            }

            if (lexer.lookAhead() == '{') {
                try {

                    body = parseJavaStatement(lexer);

                } catch (CompileError e) {
                    throw new ParseException(e);
                }
                if (!(lexer.lookAhead() == ';')) {
                    throw new ParseException("; is not found." + lexer);
                }
                lexer.get();
            }

            if (targetType != null && !name.equals("new")) {
                ASTMethod m = new ASTMethod(mod, returnType, name, param,
                        throwable, new ASTBlock(body));
                PcsType t = makePcsType(targetType);
                return new IntertypeDecl(t, m);

            } else if (targetType == null && name.equals("new")) {
                ASTConstructor cnst = new ASTConstructor(mod, param, throwable,
                        new ASTBlock(body));
                PcsType t = makePcsType(returnType);
                return new IntertypeDecl(t, cnst);
            }

        }

        // read field declaration
        // initializer
        String code = null;
        if (lexer.lookAhead() == '=') {
            lexer.get();
            try {
                lexer.startMark();
                ASTree t = new Parser(lexer.lex)
                        .parseExpression(new SymbolTable());
                code = lexer.endMark();
            } catch (CompileError e) {
                throw new ParseException(e);
            }
        }
        if (!(lexer.get() == ';')) {
            throw parseError(c, ";");
        }
        ASTField f;
        if (name.equals(ANONYMOUS_FIELD_NAME)) {
            f = new ASTAnonymousField(mod, returnType, new ASTExpr(code),
                    new ASTType[0]);
        } else {
            f = new ASTField(mod, returnType, name, new ASTExpr(code));
        }
        PcsType t = makePcsType(targetType);
        return new IntertypeDecl(t, f);
    }

    static String parseJavaStatement(Lexer l) throws CompileError {
        l.startMark();
        Stmnt t = new Parser(l.lex).parseStatement(new SymbolTable());
        String body = l.endMark();
        return body;
    }

    /*
     * Parse the signature of the type member and returns type name before the
     * last identifier. for example, if tokenBuf are
     * "java.util.List.add(Object)", this method returns the ASTType object
     * which means "java.util.List". if tokenBuf are "java.util.List", this
     * method returns the ASTType object which means "java.util.List".
     */

    private ASTType parseJavaTypeBeforeLastId(Context c) throws ParseException {

        if (lexer.isPrimitiveType()) {
            String primitive = lexer.getPrimitiveType();
            lexer.get();
            return new ASTPrimitiveType(primitive);
        }

        if (lexer.lookAhead() == Identifier) {
            StringBuffer strbuf = new StringBuffer();
            strbuf.append(lexer.getIdentifier());
            lexer.get(); //Identifier

            while (lexer.lookAhead() == '.') {
                /*
                 * ". <name>(" or ". <name>=" or " <name>;" Ȃname c
                 */
                if (lexer.lookAhead() == '.'
                        && (lexer.lookAhead(1) == Identifier || lexer
                                .lookAhead(1) == NEW)
                        && (lexer.lookAhead(2) == '='
                                || lexer.lookAhead(2) == '(' || lexer
                                .lookAhead(2) == ';')) {
                    break;
                }

                lexer.get(); // '.'
                if (lexer.lookAhead() == Identifier) {
                    strbuf.append("." + lexer.getIdentifier());
                    lexer.get(); //Identifier
                } else {
                    throw parseError(c, "Identifier");
                }
            }
            return new ASTType(strbuf.toString());
        } else
            throw parseError(c, "Identifier");
    }

    static public PcsType makePcsType(ASTType type) {

        if (type instanceof ASTPrimitiveType) {
            TypeNamePattern t = new PrimitiveTypeNamePattern(type.getType());
            return new PcsType(new PcsModifier(), t);
        }

        String str = type.getType();
        StringBuffer buf = new StringBuffer(str);

        SingleTypeNamePattern current;

        if (buf.indexOf(".") == -1) {
            current = new IdentifierTypeNamePattern(new IdPattern(buf
                    .toString()));

        } else {

            String s = buf.substring(0, buf.indexOf("."));
            buf.delete(0, buf.indexOf(".") + 1);
            IdentifierTypeNamePattern n = new IdentifierTypeNamePattern(
                    new IdPattern(s));
            current = n;
        }

        while (!(buf.indexOf(".") == -1)) {
            String s = buf.substring(0, buf.indexOf("."));
            buf.delete(0, buf.indexOf(".") + 1);
            IdentifierTypeNamePattern n = new IdentifierTypeNamePattern(
                    new IdPattern(s));
            current = new DotTypeNamePattern(current, n);
        }

        IdentifierTypeNamePattern n = new IdentifierTypeNamePattern(
                new IdPattern(buf.toString()));

        TypeNamePattern t = new DotTypeNamePattern(current, n);
        return new PcsType(new PcsModifier(), t);
    }

    /*
     * java.modifiers :=
     */
    private Modifier parseJavaModifiers(Context c) {
        Modifier m = new Modifier();
        while (true) {
            if (lexer.lookAhead() == STATIC) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.STATIC);
            } else if (lexer.lookAhead() == ABSTRACT) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.ABSTRACT);
            } else if (lexer.lookAhead() == FINAL) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.FINAL);
            } else if (lexer.lookAhead() == NATIVE) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.NATIVE);
            } else if (lexer.lookAhead() == PRIVATE) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.PRIVATE);
            } else if (lexer.lookAhead() == PROTECTED) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.PROTECTED);
            } else if (lexer.lookAhead() == PUBLIC) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.PUBLIC);
            } else if (lexer.lookAhead() == STRICT) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.STRICT);
            } else if (lexer.lookAhead() == SYNCHRONIZED) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.SYNCHRONIZED);
            } else if (lexer.lookAhead() == TRANSIENT) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.TRANSIENT);
            } else if (lexer.lookAhead() == VOLATILE) {
                lexer.get();
                m.setModifier(java.lang.reflect.Modifier.VOLATILE);
            } else {
                break;
            }
        }
        return m;
    }

    /*
     * java.type := identifier ( "." identifier)*
     */
    private ASTType parseJavaType(Context c) throws ParseException {

        if (lexer.isPrimitiveType()) {
            String primitive = lexer.getPrimitiveType();
            lexer.get();
            return new ASTPrimitiveType(primitive);
        }

        if (lexer.lookAhead() == Identifier) {
            StringBuffer strbuf = new StringBuffer();
            strbuf.append(lexer.getIdentifier());
            lexer.get(); //Identifier

            while (lexer.lookAhead() == '.') {
                lexer.get(); // '.'
                if (lexer.lookAhead() == Identifier) {
                    strbuf.append("." + lexer.getIdentifier());
                    lexer.get(); //Identifier
                } else
                    throw parseError(c, "Identifier");
            }
            return new ASTType(strbuf.toString());
        } else
            throw parseError(c, "Identifier");
    }

    /*
     * target.type := java.type | "Cflow(" pointcut.expr ")"
     */
    private ASTType parseDeclaringClassType(Context c) throws ParseException {

        if (lexer.lookAhead() == Identifier
                && lexer.getIdentifier().equals(CFLOW)) {
            return parseCflowType(c);
        } else {
            return parseJavaTypeBeforeLastId(c);
        }
    }

    private ASTCflowType parseCflowType(Context c) throws ParseException {

        lexer.get();//Identifier
        if (lexer.lookAhead() != '(') {
            throw parseError(c, ")");
        } else {
            lexer.get();//'('
        }
        ASTPointcutExpr expr;
        try {
            expr = new PointcutParser(lexer).parsePointcut();
        } catch (ParseException e) {
            throw new ParseException(
                    "Syntax error. The syntax of the pointcut parameter of Cflow at <reference> tag is wrong.",
                    e);
        }
        if (lexer.lookAhead() != ')') {
            throw parseError(c, "(");
        } else {
            lexer.get();//'('
        }
        return new ASTCflowType(expr);
    }

    /*
     * parameter := "(" injected.type ("," injected.type )* ")"
     */
    private ASTFormalParameter[] parseParameters(Context c)
            throws ParseException {

        if (lexer.lookAhead() != '(') {
            throw parseError(c, ")");
        } else {
            lexer.get();//'('
        }

        List typeList = new LinkedList();

        if (lexer.lookAhead() == TokenId.Identifier || lexer.isPrimitiveType()) {
            typeList.add(parseParameter(c));
            while (lexer.lookAhead() == ',') {
                lexer.get();
                typeList.add(parseParameter(c));
            }
        }
        if (lexer.lookAhead() != ')') {
            throw parseError(c, "(");
        } else {
            lexer.get();//'('
        }
        return (ASTFormalParameter[]) typeList
                .toArray(new ASTFormalParameter[typeList.size()]);
    }

    private ASTFormalParameter parseParameter(Context c) throws ParseException {

        ASTType type = parseJavaType(c);
        if (lexer.lookAhead() == Identifier) {
            String name = lexer.getIdentifier();
            lexer.get();
            return new ASTFormalParameter(type, name);
        } else {
            return new ASTFormalParameter(type, null);
        }

    }

    private ASTType[] parseTypes(Context c) throws ParseException {
        List typeList = new LinkedList();

        if (lexer.lookAhead() == TokenId.Identifier) {
            typeList.add(parseJavaType(c));
            while (lexer.lookAhead() == ',') {
                lexer.get();
                typeList.add(parseJavaType(c));
            }
        }
        return (ASTType[]) typeList.toArray(new ASTType[typeList.size()]);
    }

    private ParseException parseError(Context c, String string) {
        return new ParseException("\"" + string + "\" expected. lexer dump:"
                + lexer);
    }

    private class Context {

    }

}