/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.parser;

import gluonj.ast.java.ASTFormalParameter;
import gluonj.ast.java.ASTType;
import gluonj.ast.pattern.AndTypeNamePattern;
import gluonj.ast.pattern.ArrayTypeNamePattern;
import gluonj.ast.pattern.BasicTypeNamePattern;
import gluonj.ast.pattern.DotDotTypeNamePattern;
import gluonj.ast.pattern.DotTypeNamePattern;
import gluonj.ast.pattern.IdPattern;
import gluonj.ast.pattern.IdentifierTypeNamePattern;
import gluonj.ast.pattern.NotTypeNamePattern;
import gluonj.ast.pattern.OrTypeNamePattern;
import gluonj.ast.pattern.PrimitiveTypeNamePattern;
import gluonj.ast.pattern.SingleTypeNamePattern;
import gluonj.ast.pattern.SubTypeNamePattern;
import gluonj.ast.pattern.TypeNamePattern;
import gluonj.ast.pcs.PcsBehavior;
import gluonj.ast.pcs.PcsConstructor;
import gluonj.ast.pcs.PcsField;
import gluonj.ast.pcs.PcsMethod;
import gluonj.ast.pcs.PcsModifier;
import gluonj.ast.pcs.PcsParameter;
import gluonj.ast.pcs.PcsParameterDotDot;
import gluonj.ast.pcs.PcsParameterType;
import gluonj.ast.pcs.PcsType;
import gluonj.ast.pcs.PointcutShadow;
import gluonj.ast.pcs.Residue;
import gluonj.ast.pointcut.ASTAnd;
import gluonj.ast.pointcut.ASTArgs;
import gluonj.ast.pointcut.ASTCflow;
import gluonj.ast.pointcut.ASTCflowBelow;
import gluonj.ast.pointcut.ASTFieldGet;
import gluonj.ast.pointcut.ASTFieldSet;
import gluonj.ast.pointcut.ASTMethodCall;
import gluonj.ast.pointcut.ASTMethodExecution;
import gluonj.ast.pointcut.ASTNot;
import gluonj.ast.pointcut.ASTOr;
import gluonj.ast.pointcut.ASTPCReference;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.ast.pointcut.ASTPointcutPrimitive;
import gluonj.ast.pointcut.ASTResidue;
import gluonj.ast.pointcut.ASTTarget;
import gluonj.ast.pointcut.ASTThis;
import gluonj.ast.pointcut.ASTWithin;
import gluonj.ast.pointcut.ASTWithincode;
import gluonj.weaver.JoinPointConstants;

import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javassist.compiler.CompileError;

public class PointcutParser extends AbstractPointcutParser implements TokenId,
        JoinPointConstants {

    public static final int ID = 1;

    public static final int POINTCUT = 2;

    public static final int JAVA_EXPR = 4;

    public static final int JAVA_CONST = 8;

    public static final int PATTERN = 16;

    public static final int CLASS_PATTERN = 32;

    public static final int FIELD_PATTERN = 64;

    public static final int BEHAVIOR_PATTERN = 128;

    private Lexer lexer;

    public PointcutParser(Lexer lex) {
        lexer = lex;
    }

    /*
     * Tokens used in this class : ( ) || && ! , . * + IDENTIFIER THROWS
     */

    public ASTPointcutExpr parsePointcut() throws ParseException {

        return parsePointcut(new Context());
    }

    public ASTPointcutExpr parsePointcut(AbstractPointcutParser.Context c)
            throws ParseException {

        return parsePointcut((Context) c);
    }

    /*
     * pointcut := pointcut.or
     */

    public ASTPointcutExpr parsePointcut(Context c) throws ParseException {

        return parsePointcutOr(c);
    }

    /*
     * pointcut.or := pointcut.and (("||" | "OR" | "OROR" ) pointcut.and)*
     */

    private ASTPointcutExpr parsePointcutOr(Context c) throws ParseException {

        ASTPointcutExpr e = parsePointcutAnd(c);

        while (lexer.lookAhead() == OROR
                || (lexer.lookAhead() == Identifier && (lexer.getIdentifier()
                        .equals("OROR") || lexer.getIdentifier().equals("OR")))) {

            lexer.get();
            e = new ASTOr(e, parsePointcutAnd(c));
        }
        return e;
    }

    /*
     * pointcut.and := pointcut.not (("&&" | "ANDAND" ) pointcut.not)*
     */

    private ASTPointcutExpr parsePointcutAnd(Context c) throws ParseException {

        ASTPointcutExpr e = parsePointcutNot(c);

        while (lexer.lookAhead() == ANDAND
                || (lexer.lookAhead() == Identifier && (lexer.getIdentifier()
                        .equals("ANDAND") || lexer.getIdentifier()
                        .equals("AND")))) {
            lexer.get();
            e = new ASTAnd(e, parsePointcutNot(c));
        }
        return e;
    }

    /*
     * pointcut.not := ["!"]? pointcut.unary
     */
    private ASTPointcutExpr parsePointcutNot(Context c) throws ParseException {

        if (lexer.lookAhead() == '!') {
            lexer.get();
            return new ASTNot(parsePointcutUnary(c));
        } else {
            return parsePointcutUnary(c);
        }
    }

    /*
     * pointcut.unary := "(" pointcut ")" | pointcut.designater
     */
    private ASTPointcutExpr parsePointcutUnary(Context c) throws ParseException {

        if (lexer.lookAhead() == '(') {
            lexer.get();
            ASTPointcutExpr expr = parsePointcut(c);
            if (lexer.lookAhead() == ')') {
                lexer.get();
                return expr;
            } else
                throw parseError(c, lexer.lookAhead(), new int[] { ')' });
        } else {
            return parsePointcutDesignator(c);
        }
    }

    /*
     * pointcut.designater := (pointcut.identifier | "this") "("
     * pointcut.arguments ")"
     */

    private ASTPointcutExpr parsePointcutDesignator(Context c)
            throws ParseException {

        String id;
        if (lexer.lookAhead() == TokenId.THIS) {
            lexer.get();
            id = "this";
        } else {
            id = parsePointcutIdentifier(c);
        }
        int expected;
        if (lexer.lookAhead() == '(') {
            lexer.get();

            c.pushArgumentType(getArgumentType(id));
            PointcutArgument[] args = parsePointcutArguments(c);
            c.popArgumentType();

            if (lexer.lookAhead() == ')') {
                lexer.get();
                ASTPointcutExpr expr = makeDesignator(id, args, c);
                return expr;
            } else {
                expected = ')';
            }
        } else {
            expected = '(';
        }
        throw parseError(c, lexer.lookAhead(), new int[] { expected });
    }

    private int getArgumentType(String pcIdentifier) {

        if (pcIdentifier.equals("call")) {
            return BEHAVIOR_PATTERN;
        }
        if (pcIdentifier.equals("execution")) {
            return BEHAVIOR_PATTERN;
        }
        if (pcIdentifier.equals("set")) {
            return FIELD_PATTERN;
        }
        if (pcIdentifier.equals("get")) {
            return FIELD_PATTERN;
        }
        if (pcIdentifier.equals("within")) {
            return CLASS_PATTERN;
        }
        if (pcIdentifier.equals("withincode")) {
            return BEHAVIOR_PATTERN;
        }
        if (pcIdentifier.equals("target")) {
            return ID;
        }
        if (pcIdentifier.equals("this")) {
            return ID;
        }
        if (pcIdentifier.equals("args")) {
            return ID;
        }
        if (pcIdentifier.equals("cflow")) {
            return POINTCUT;
        }
        if (pcIdentifier.equals("cflowbelow")) {
            return POINTCUT;
        }
        if (pcIdentifier.equals("if")) {
            return JAVA_EXPR;
        }
        //if(pcIdentifier.equals("")) {
        //}
        return JAVA_CONST;
    }

    /*
     * pointcut.identifier := IDENTIFIER
     */

    private String parsePointcutIdentifier(Context c) throws ParseException {
        return parseIdentifier(c);
    }

    /*
     * pointcut.arguments := (pointcut.argument ("," pointcut.argument)* )?
     */

    private PointcutArgument[] parsePointcutArguments(Context c)
            throws ParseException {

        if (lexer.lookAhead() == ')') {
            return new PointcutArgument[0];
        }
        List argsList = new LinkedList();
        argsList.add(parsePointcutArgument(c));
        while (lexer.lookAhead() == ',') {
            lexer.get();
            argsList.add(parsePointcutArgument(c));
        }
        PointcutArgument[] args = new PointcutArgument[argsList.size()];
        argsList.toArray(args);
        return args;
    }

    /*
     * pointcut.argument :=
     * 
     * java.expression // for if | pointcut // for cflow | java.variant // for
     * binding | java.type // for condition type  variant ̋ʂ? |
     * pointcut.pattern | java.constant //
     * 
     * 
     * IDENTIFIER + ")" -> type or variant IDENTIFIER + "(" -> pc? possible to
     * if modifier -> pattern
     *  
     */

    private PointcutArgument parsePointcutArgument(Context c)
            throws ParseException {

        /*
         * pointcut.argument := java.variantortype | pointcut | java.expression |
         * java.constant | pointcut.pattern.class | pointcut.pattern.field |
         * pointcut.pattern.behavior
         *  
         */

        if (c.matchArgumentType(ID)) {
            // target, args, this
            return PointcutArgument.make(parseVariantOrType(c));

        }
        if (c.matchArgumentType(POINTCUT)) {
            // cflow
            return PointcutArgument.make(parsePointcut(c));
        }
        if (c.matchArgumentType(JAVA_EXPR)) {
            try {
                return PointcutArgument.make(InjectionCodeParser
                        .parseJavaStatement(lexer));
            } catch (CompileError e) {
                throw new ParseException(e);
            }
        }
        if (c.matchArgumentType(JAVA_CONST)) {
            // named pointcut
            if (lexer.lookAhead() == Identifier) {
                return PointcutArgument.make(parseVariable(c));
            } else {
                return PointcutArgument.make(parseJavaConstrant(c));
            }
        }
        if (c.matchArgumentType(CLASS_PATTERN)) {
            // within, staticinitialization
            return PointcutArgument.make(parsePatternType(c));
        }
        if (c.matchArgumentType(FIELD_PATTERN)) {
            // set, get
            return PointcutArgument.make(parseFieldPattern(c));
        }
        if (c.matchArgumentType(BEHAVIOR_PATTERN)) {
            // execution, call, withincode
            return PointcutArgument.make(parseBehaviorPattern(c));
        }
        throw parseError(c, lexer.lookAhead(), new int[] {});
    }

    /*
     * pointcut.pattern.behavior := | pattern.modifier pattern.type pattern.type
     * "(" pattern.args ")" pattern.throws | pattern.modifier pattern.type "("
     * pattern.args ")" pattern.throws
     */

    private PcsBehavior parseBehaviorPattern(Context c) throws ParseException {

        //parse modifiers
        PcsModifier modifiers = parsePatternModifier(c);
        lexer.toString();

        //parse typePattern
        PcsType t = parsePatternTypeInPCArgs(c);
        PcsType t2 = null;

        readSeparetor(c);

        if (lexer.lookAhead() == Identifier || lexer.lookAhead() == '*') {
            t2 = parsePatternTypeInPCArgs(c);
        }
        //pass '.', '..'
        if (lexer.lookAhead() == '.') {
            lexer.get();
        }
        if (lexer.lookAhead() == '.') {
            lexer.get();
        }
        IdPattern name = null;
        boolean isNameNew = false;
        PcsParameter[] args = null;
        PcsType[] throwables = null;

        if (lexer.lookAhead() == NEW) {
            lexer.get(); //new
            isNameNew = true;
        } else {
            name = parsePatternId(c);
        }

        if (lexer.lookAhead() == '(') {

            lexer.get(); // "("
            args = parsePatternArgument(c);
            if (lexer.lookAhead() == ')') {
                lexer.get(); // ")"
                throwables = parsePatternThrows(c);
            }
        }
        if (args == null || throwables == null) {
            throw parseError(c, lexer.lookAhead(), new int[] {});
        }

        /* constructor pattern */
        if (isNameNew && t2 == null) {
            return new PcsConstructor(modifiers, t, args, throwables);

            /* method pattern */
        } else if (!isNameNew && t2 != null) {
            return new PcsMethod(modifiers, t, t2, name, args, throwables);
        }
        throw new ParseException("Constructor or Method signature is wrong.");
    }

    /*
     * pointcut.pattern.field := pattern.modifier pattern.type pattern.type
     */

    private PcsField parseFieldPattern(Context c) throws ParseException {

        PcsModifier modifiers = parsePatternModifier(c);
        PcsType t = parsePatternType(c);

        readSeparetor(c);

        PcsType t2 = parsePatternTypeInPCArgs(c);

        //pass '.', '..'
        if (lexer.lookAhead() == '.') {
            lexer.get();
        }
        if (lexer.lookAhead() == '.') {
            lexer.get();
        }

        IdPattern id = parsePatternId(c);

        return new PcsField(modifiers, t, t2, id);
    }

    /*
     * pointcut.pattern.class := pattern.modifier pattern.type
     */

    /*
     * private TypePattern parseClassPattern(Context c) throws ParseException {
     * 
     * int modifiers = parsePatternModifier(c); TypePattern t =
     * parsePatternType(c);
     * 
     * return makeClassPattern(modifiers, t); }
     */

    /*
     * separator := '%'?
     */

    private void readSeparetor(Context c) {
        if (lexer.lookAhead() == '%')
            lexer.get();
    }

    /*
     * java.variable := Identifier
     */
    private String parseVariable(Context c) {
        String var = lexer.getIdentifier();
        //CHECK
        lexer.get(); //Identifier
        return var;
    }

    /*
     * java.constant :=
     *  
     */
    private String parseJavaConstrant(Context c) throws ParseException {

        if (lexer.lookAhead() == IntConstant
                || lexer.lookAhead() == LongConstant) {
            return new Long(lexer.getLong()).toString();
        } else if (lexer.lookAhead() == FloatConstant
                || lexer.lookAhead() == DoubleConstant) {
            return new Double(lexer.getDouble()).toString();
        } else if (lexer.lookAhead() == StringL) {
            return lexer.getString();
        }
        throw parseError(c, lexer.lookAhead(), new int[] {});
    }

    /*
     * java.variantortype := ".." | Identifier ("." identifier)* |
     * java.primitive
     */
    private String parseVariantOrType(Context c) throws ParseException {

        if (lexer.lookAhead() == '.' && lexer.lookAhead(1) == '.') {
            lexer.get();
            lexer.get();
            return "..";
        }

        if (lexer.lookAhead() == Identifier) {

            StringBuffer str = new StringBuffer(lexer.getIdentifier());
            //CHECK
            lexer.get(); //Identifier

            while (lexer.lookAhead() == '.' && lexer.lookAhead(1) == Identifier) {
                lexer.get();
                lexer.lookAhead();
                str.append("." + lexer.getIdentifier());
                //CHECK
                lexer.get(); //Identifier
            }
            return str.toString();
        }
        if (lexer.isPrimitiveType()) {
            return lexer.getPrimitiveType();
        }

        throw parseError(c, lexer.lookAhead(), new int[] {});
    }

    /*
     * pattern.modifier := ("!"? ("static" || "public" || "private" || ... ) )*
     */

    private PcsModifier parsePatternModifier(Context c) throws ParseException {
        PcsModifier p = new PcsModifier();

        while (true) {
            int i = lexer.lookAhead();
            boolean has = true;
            if (i == '!') {
                has = false;
                lexer.get();
                i = lexer.lookAhead();
            }
            if (i == PUBLIC) {
                p.setModifiers(Modifier.PUBLIC, has);
            } else if (i == PRIVATE) {
                p.setModifiers(Modifier.PRIVATE, has);
            } else if (i == PROTECTED) {
                p.setModifiers(Modifier.PROTECTED, has);
            } else if (i == STATIC) {
                p.setModifiers(Modifier.STATIC, has);
            } else if (i == FINAL) {
                p.setModifiers(Modifier.FINAL, has);
            } else if (i == SYNCHRONIZED) {
                p.setModifiers(Modifier.SYNCHRONIZED, has);
            } else if (i == VOLATILE) {
                p.setModifiers(Modifier.VOLATILE, has);
            } else if (i == TRANSIENT) {
                p.setModifiers(Modifier.TRANSIENT, has);
            } else if (i == NATIVE) {
                p.setModifiers(Modifier.NATIVE, has);
            } else if (i == INTERFACE) {
                p.setModifiers(Modifier.INTERFACE, has);
            } else if (i == ABSTRACT) {
                p.setModifiers(Modifier.ABSTRACT, has);
            } else if (i == STRICT) {
                p.setModifiers(Modifier.STRICT, has);
            } else {
                if (!has) {
                    throw new ParseException("!");
                } else {
                    break;
                }
            }
            lexer.get();
        }
        return p;
    }

    /*
     * pattern.type := pattern.id ((".." | ".") pattern.id)* [+]? |
     * java.primitive
     */

    private PcsType parsePatternType(Context c) throws ParseException {

        PcsType t = new PcsType(new PcsModifier(), parsePatternTypeOr(c));
        return t;

    }

    // typePat1 typePat2.idPat( -> typePat1 ends ID && typePad2 ends ID
    // typePat1 typePat2( -> typePat1 ends ID && typePad2 ends ID
    // typePat.idPat( -> typePatter contains "&&" or "||"
    // typePat( -> typePattern does not contains "&&" and "||"

    /*
     * typeNamePat ǂݍނƂɁA ǂސ ("." || "..") IDPat "(" Ƃp^[΁A IDPat
     * ̎O܂łǂ܂Ȃ
     * 
     * : typePat..idPat -> (typePat..*).idPat
     */
    private PcsType parsePatternTypeInPCArgs(Context c) throws ParseException {
        c.setReadingDeclaringClass(true);
        PcsType t = new PcsType(new PcsModifier(), parsePatternTypeOr(c));
        c.setReadingDeclaringClass(false);
        return t;
    }

    /*
     * pattern.type.or := pattern.type.and (("||" | "OROR") pattern.type.and )*
     */
    private TypeNamePattern parsePatternTypeOr(Context c) throws ParseException {
        TypeNamePattern t = parsePatternTypeAnd(c);
        while (lexer.lookAhead() == OROR
                || (lexer.lookAhead() == Identifier && lexer.getIdentifier()
                        .equals("OROR"))) {
            lexer.get();
            t = new OrTypeNamePattern(t, parsePatternTypeAnd(c));
        }
        return t;
    }

    /*
     * pattern.type.and := pattern.type.not (("&&" | "ANDAND") pattern.type.not )*
     */
    private TypeNamePattern parsePatternTypeAnd(Context c)
            throws ParseException {
        TypeNamePattern t = parsePatternTypeNot(c);
        while (lexer.lookAhead() == ANDAND
                || (lexer.lookAhead() == Identifier && lexer.getIdentifier()
                        .equals("ANDAND"))) {
            lexer.get();
            t = new AndTypeNamePattern(t, parsePatternTypeNot(c));
        }
        return t;
    }

    /*
     * pattern.type.not := "!"? pattern.type.unary
     */
    private TypeNamePattern parsePatternTypeNot(Context c)
            throws ParseException {
        if (lexer.lookAhead() == '!') {
            lexer.get();
            return new NotTypeNamePattern(parsePatternTypeUnary(c));
        } else {
            return parsePatternTypeUnary(c);
        }

    }

    /*
     * pattern.type.unary := "(" pattern.type.or ")" | pattern.type.subtype
     */
    private TypeNamePattern parsePatternTypeUnary(Context c)
            throws ParseException {

        if (lexer.lookAhead() == '(') {
            lexer.get();

            //() ̒ǂłꍇAidPat ܂œǂ݂Ă܂
            // Sz͂Ȃ
            // Ex. (String || Test.Inner).id()
            c.setReadingDeclaringClass(false);
            TypeNamePattern t = parsePatternTypeOr(c);
            c.setReadingDeclaringClass(true);

            if (lexer.lookAhead() == ')') {
                lexer.get();
                return t;
            } else
                throw parseError(c, lexer.lookAhead(), new int[] { ')' });
        } else {
            BasicTypeNamePattern[] p = parsePatternTypeSubtype(c);
            TypeNamePattern t = p[0];
            for (int i = 1; i < p.length; i++) {
                t = new OrTypeNamePattern(t, p[i]);
            }
            return t;
        }
    }

    /*
     * pattern.type.subtype := pattern.type.array "+"?
     */

    private BasicTypeNamePattern[] parsePatternTypeSubtype(Context c)
            throws ParseException {
        BasicTypeNamePattern[] t = parsePatternTypeArray(c);
        if (lexer.lookAhead() == '+') {
            lexer.get();//+
            for (int i = 0; i < t.length; i++) {
                t[i] = new SubTypeNamePattern(t[i]);
            }
            return t;
        } else {
            return t;
        }
    }

    /*
     * pattern.type.array := pattern.type.single "[]"*
     */

    private BasicTypeNamePattern[] parsePatternTypeArray(Context c)
            throws ParseException {
        BasicTypeNamePattern[] t = parsePatternTypeSingle(c);

        while (lexer.lookAhead() == '[' && lexer.lookAhead(1) == ']') {
            lexer.get();
            lexer.get();
            for (int i = 0; i < t.length; i++) {
                t[i] = new ArrayTypeNamePattern(t[i]);
            }
        }
        return t;
    }

    /*
     * pattern.type.single := "boolean" | "char" ... | "void" |
     * pattern.type.dotordotdot
     */
    private BasicTypeNamePattern[] parsePatternTypeSingle(Context c)
            throws ParseException {
        if (lexer.isPrimitiveType()) {
            String s = lexer.getPrimitiveType();
            lexer.get();//primitive
            return new BasicTypeNamePattern[] { new PrimitiveTypeNamePattern(s) };
        } else {
            SingleTypeNamePattern p = parsePatternDotOrDotDot(c);
            String[] pck = c.getImportedPackage();
            if (pck.length == 0 || p.getInString().indexOf(".") != -1) {
                return new BasicTypeNamePattern[] { p };
            } else {
                BasicTypeNamePattern[] ps = new BasicTypeNamePattern[pck.length];
                for (int i = 0; i < ps.length; i++) {
                    ps[i] = new SingleTypeNamePattern(c.getImportedPackage()[i]
                            + "." + p.getInString());
                }
                return ps;
            }
        }
    }

    /*
     * pattern.type.dotordotdot := pattern.id ((".." | ".") pattern.id)*
     */
    private SingleTypeNamePattern parsePatternDotOrDotDot(Context c)
            throws ParseException {

        SingleTypeNamePattern i = new IdentifierTypeNamePattern(
                parsePatternId(c));

        while (true) {

            /*
             * next tokenBuf  (".." | ".") idPat1 ("(" | ")") Ƃ`ĂA ̕
             * \bhnew \B̂".." | "." ȍ~ParseȂ
             */
            if (c.readingDeclaringClass()) {
                if (lexer.lookAhead() == '.' && lexer.lookAhead(1) == '.') {
                    int idx = isBehaviorName(2);
                    if (idx != -1
                            && (lexer.lookAhead(idx) == '(' || lexer
                                    .lookAhead(idx) == ')')) {
                        i = new DotDotTypeNamePattern(i,
                                new IdentifierTypeNamePattern(
                                        new IdPattern("*")));
                        return i;
                    }
                } else if (lexer.lookAhead() == '.') {
                    int idx = isBehaviorName(1);
                    if (idx != -1
                            && (lexer.lookAhead(idx) == '(' || lexer
                                    .lookAhead(idx) == ')')) {
                        return i;
                    }
                }
            }

            if (lexer.lookAhead() == '.' && lexer.lookAhead(1) == '.') {
                lexer.get();
                lexer.get();
                i = new DotDotTypeNamePattern(i, new IdentifierTypeNamePattern(
                        parsePatternId(c)));

            } else if (lexer.lookAhead() == '.') {
                lexer.get();
                i = new DotTypeNamePattern(i, new IdentifierTypeNamePattern(
                        parsePatternId(c)));
            } else {
                break;
            }
        }
        return i;
    }

    private int isBehaviorName(int idx) {

        int top_idx = idx;

        if (lexer.lookAhead(idx) == NEW) {
            return ++idx;
        }

        boolean readID = false;
        while ((lexer.lookAhead(idx) == Identifier && !readID)
                || lexer.lookAhead(idx) == '*') {
            if (lexer.lookAhead(idx) == Identifier) {
                readID = true;
            } else {
                readID = false;
            }
            idx++;
        }
        if (idx > top_idx) {
            return idx;
        } else {
            return -1;
        }
    }

    /*
     * pattern.id := (pattern.name | "*")*
     */

    private IdPattern parsePatternId(Context c) throws ParseException {

        boolean readID = false;
        StringBuffer buf = new StringBuffer();

        while (lexer.lookAhead() == '*'
                || (lexer.lookAhead() == Identifier && !readID)) {

            if (lexer.lookAhead() == '*') {
                lexer.get();
                readID = false;
                buf.append("*");
            } else {
                buf.append(lexer.getIdentifier());
                lexer.get();
                readID = true;
            }
        }
        IdPattern i = new IdPattern(buf.toString());
        return i;
    }

    /*
     * pattern.args := [(".." | pattern.type) ( "," (".." | pattern.type))* ]?
     *  
     */

    private PcsParameter[] parsePatternArgument(Context c)
            throws ParseException {

        if (lexer.lookAhead() == ')') {
            return new PcsParameter[0];
        }

        List argsList = new LinkedList();

        PcsParameter p;
        p = parsePatternParameterElem(c);
        argsList.add(p);

        while (lexer.lookAhead() == ',') {
            lexer.get();
            p = parsePatternParameterElem(c);
            argsList.add(p);
        }
        PcsParameter[] param = new PcsParameter[argsList.size()];
        argsList.toArray(param);
        return param;
    }

    private PcsParameter parsePatternParameterElem(Context c)
            throws ParseException {
        PcsParameter p;
        if (lexer.lookAhead() == '.' && lexer.lookAhead(1) == '.') {
            lexer.get();
            lexer.get();
            p = new PcsParameterDotDot();
        } else {
            PcsType t = parsePatternType(c);
            p = new PcsParameterType(t);
        }
        return p;
    }

    /*
     * pattern.throws := ["throws" pattern.type ("," pattern.type)* ]?
     */

    private PcsType[] parsePatternThrows(Context c) throws ParseException {

        if (lexer.lookAhead() == ')')
            return new PcsType[0];

        if (lexer.lookAhead() != THROWS) {
            throw parseError(c, lexer.lookAhead(), new int[] { THROWS });
        }
        lexer.get();
        List argsList = new LinkedList();
        PcsType t = parsePatternType(c);
        argsList.add(t);

        while (lexer.lookAhead() == ',') {
            lexer.get();
            t = parsePatternType(c);
            argsList.add(t);
        }
        PcsType[] args = new PcsType[argsList.size()];
        argsList.toArray(args);
        return args;
    }

    /*
     * pattern.name := IDENTIFIER
     */

    private String parseJavaName(Context c) throws ParseException {

        if (lexer.lookAhead() == Identifier) {
            String id = lexer.getIdentifier();
            //CHECK
            lexer.get(); //Identifier
            return id;
        }
        throw parseError(c, lexer.lookAhead(), new int[] { Identifier });
    }

    /*
     * java.IDENTIFIER
     */

    private String parseIdentifier(Context c) throws ParseException {
        if (lexer.lookAhead() == Identifier) {
            String id = lexer.getIdentifier();
            //CHECK
            lexer.get(); //Identifier
            return id;
        }
        throw parseError(c, lexer.lookAhead(), new int[] { Identifier });
    }

    private ASTPointcutExpr makeDesignator(String id, PointcutArgument[] args,
            Context c) throws ParseException {

        if (designatorNameToInt(id) == JoinPointConstants.CALL) {
            if (args.length == 1 && args[0].type == PATTERN) {

                return new ASTMethodCall(args[0].behaviorPattern);
            }
        } else if (designatorNameToInt(id) == EXECUTION) {
            if (args.length == 1 && args[0].type == PATTERN) {

                return new ASTMethodExecution(args[0].behaviorPattern);
            }
        } else if (designatorNameToInt(id) == FIELD_SET) {
            if (args.length == 1 && args[0].type == PATTERN) {
                return new ASTFieldSet(args[0].fieldPattern);
            }
        } else if (designatorNameToInt(id) == FIELD_GET) {
            if (args.length == 1) {
                return new ASTFieldGet(args[0].fieldPattern);
            }
        } else if (designatorNameToInt(id) == WITHIN) {
            if (args.length == 1 && args[0].type == PATTERN) {
                return new ASTWithin(args[0].classPattern);
            }
        } else if (designatorNameToInt(id) == WITHINCODE) {
            if (args.length == 1 && args[0].type == PATTERN) {
                return new ASTWithincode(args[0].behaviorPattern);
            }
        } else if (designatorNameToInt(id) == TARGET) {
            if (args.length == 1 && args[0].type == ID
                    && !args[0].variantOrType.equals("..")) {

                ASTTarget t;
                String name = args[0].variantOrType;
                if (c.isPCParameterName(name)) {
                    ASTType type = c.getPCParameterType(name);
                    t = new ASTTarget(name, type.getType());
                } else {
                    t = new ASTTarget(name);
                }
                return t;
            }

        } else if (designatorNameToInt(id) == JoinPointConstants.THIS) {
            if (args.length == 1 && args[0].type == ID
                    && !args[0].variantOrType.equals("..")) {

                ASTThis t;
                String name = args[0].variantOrType;
                if (c.isPCParameterName(name)) {
                    ASTType type = c.getPCParameterType(name);
                    t = new ASTThis(name, type.getType());
                } else {
                    t = new ASTThis(name);
                }
                return t;
            }
        } else if (designatorNameToInt(id) == ARGS) {
            if (args[0].type == ID) {

                ASTArgs t;
                String[] varName = new String[args.length];
                String[] typeName = new String[args.length];
                boolean isBinder = false;

                for (int i = 0; i < args.length; i++) {
                    String n = args[i].variantOrType;
                    if (c.isPCParameterName(n)) {
                        ASTType type = c.getPCParameterType(n);
                        varName[i] = n;
                        typeName[i] = type.getType();
                        isBinder = true;
                    } else {
                        typeName[i] = n;
                    }
                }
                if (isBinder) {
                    t = new ASTArgs(varName, typeName);
                } else {
                    t = new ASTArgs(typeName);
                }
                return t;
            }

        } else if (designatorNameToInt(id) == CFLOW) {
            if (args.length == 1 && args[0].type == POINTCUT) {

                return new ASTCflow(args[0].expr);
            }
        } else if (designatorNameToInt(id) == CFLOW_BELOW) {
            if (args.length == 1 && args[0].type == POINTCUT) {

                return new ASTCflowBelow(args[0].expr);
            }
        } else if (designatorNameToInt(id) == IF) {
            if (args.length == 1 && args[0].type == ID) {

                Residue r = new Residue();
                r.addDynamicCondition(args[0].variantOrType);
                return new ASTResidue(r);
            }
        } else {
            // pc reference.
            if (args.length == 0 || args[0].type == CONST) {
                String[] pc_args = new String[args.length];
                for (int i = 0; i < pc_args.length; i++) {
                    pc_args[i] = args[i].variantOrType;
                }
                return new ASTPCReference(id, pc_args);
            }
        }

        throw parseError(id + " : wrong arguments. ");
    }

    private int designatorNameToInt(String id) {

        if (id.equals("call"))
            return JoinPointConstants.CALL;
        if (id.equals("execution"))
            return EXECUTION;
        if (id.equals("set"))
            return FIELD_SET;
        if (id.equals("get"))
            return FIELD_GET;
        if (id.equals("within"))
            return WITHIN;
        if (id.equals("withincode"))
            return WITHINCODE;
        if (id.equals("target"))
            return TARGET;
        if (id.equals("this"))
            return JoinPointConstants.THIS;
        if (id.equals("args"))
            return ARGS;
        if (id.equals("cflow"))
            return CFLOW;
        if (id.equals("cflowbelow"))
            return CFLOW_BELOW;
        //		if(id.equals("")) return ;

        return REFERENCE;
    }

    private String idToString(int tokenId) {
        return "\"" + ((char) tokenId) + "\"";
    }

    private String idToString(int[] tokenIds) {
        StringBuffer buf = new StringBuffer();

        if (tokenIds.length > 0) {
            buf.append(idToString(0));
        }

        for (int i = 1; i < tokenIds.length; i++) {
            buf.append(",");
            buf.append(idToString(i));
        }
        return buf.toString();
    }

    private ParseException parseError(Context c, int i, int[] cs) {
        return new ParseException(" Syntax error on tokenBuf :  " + lexer
                + " \n  " + idToString(cs) + " are expected.");
    }

    private ParseException parseError(String s) {
        return new ParseException(s);
    }

    static class Context extends AbstractPointcutParser.Context {

        private LinkedList argumentTypeStack = new LinkedList();

        private Map pcParameterStack = new HashMap();

        private boolean readingDeclClass;

        void pushArgumentType(int i) {
            argumentTypeStack.add(0, new Integer(i));
        }

        void popArgumentType() {
            argumentTypeStack.remove(0);
        }

        boolean matchArgumentType(int i) {
            return (((Integer) argumentTypeStack.get(0)).intValue() & i) == i;
        }

        public void setReadingDeclaringClass(boolean b) {
            readingDeclClass = b;
        }

        public boolean readingDeclaringClass() {
            return readingDeclClass;
        }

        public void addPointcutParam(ASTFormalParameter param) {
            pcParameterStack.put(param.getName(), param.getType());
        }

        public void setPCParameterName(ASTType type, String var) {
            pcParameterStack.put(var, type);
        }

        public boolean isPCParameterName(String name) {
            return pcParameterStack.containsKey(name);
        }

        public ASTType getPCParameterType(String name) {
            return (ASTType) pcParameterStack.get(name);
        }

        String[] packageName = new String[0];

        public String[] getImportedPackage() {
            return packageName;
        }

        public void clearImportedPackage() {
            packageName = new String[0];
        }

        public void setImportedPackage(String[] pck) {
            packageName = pck;
        }
    }

    public PcdFactoryTable initTable() {
        return null;
    }

}

class PcdTable {

    Map pcdMap = new HashMap();

    void putPcd(String name, PcdInfo info) {
        pcdMap.put(name, info);
    }

    PcdInfo getPcdInfo(String name) {
        PcdInfo o = (PcdInfo) pcdMap.get(name);
        return o;
    }
}

abstract class PcdInfo {

    public PcdInfo(String n, int p) {
        name = n;
        paramType = p;
    }

    String name;

    int paramType;

    public abstract ASTPointcutPrimitive pcdFactory(PointcutArgument[] pcargs)
            throws ParseException;

    public String getName() {
        return name;
    }

    public int getParameterType() {
        return paramType;
    }
}

class PointcutArgument {

    int type;

    String variantOrType;

    PointcutShadow pattern;

    PcsType classPattern;

    PcsField fieldPattern;

    PcsBehavior behaviorPattern;

    ASTPointcutExpr expr;

    public static PointcutArgument make(String var) {

        PointcutArgument a = new PointcutArgument();
        a.type = PointcutParser.ID;
        a.variantOrType = var;
        return a;
    }

    public static PointcutArgument make(PointcutShadow pat) {

        PointcutArgument a = new PointcutArgument();
        a.type = PointcutParser.PATTERN;
        a.pattern = pat;
        return a;
    }

    public static PointcutArgument make(PcsType pat) {

        PointcutArgument a = new PointcutArgument();
        a.type = PointcutParser.PATTERN;
        a.classPattern = pat;
        return a;
    }

    public static PointcutArgument make(PcsField pat) {

        PointcutArgument a = new PointcutArgument();
        a.type = PointcutParser.PATTERN;
        a.fieldPattern = pat;
        return a;
    }

    public static PointcutArgument make(PcsBehavior pat) {

        PointcutArgument a = new PointcutArgument();
        a.type = PointcutParser.PATTERN;
        a.behaviorPattern = pat;
        return a;
    }

    public static PointcutArgument make(ASTPointcutExpr expr) {

        PointcutArgument a = new PointcutArgument();
        a.type = PointcutParser.POINTCUT;
        a.expr = expr;
        return a;
    }

}