/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.runtime;

import java.lang.reflect.Field;

public class Aspect {

    /**
     * Substitutes <tt>aspect</tt> object to the anonymous field of
     * <tt>target</tt> object. <tt>clazz</tt> object specifies the type of
     * anonymous field.
     */
    public static void add(Object target, Object aspect, Class clazz) {

        Field f = getAF(target.getClass(), clazz);
        try {
            f.set(target, aspect);
        } catch (IllegalArgumentException e) {
            throw new GluonJRuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new GluonJRuntimeException(e);
        }
    }

    /**
     * Returns the object which is referred to by the anonymous field of
     * <tt>target</tt> object. <tt>clazz</tt> object specifies the type of
     * anonymous field.
     */
    public static Object get(Object target, Class clazz) {
        Field f = getAF(target.getClass(), clazz);
        try {
            return f.get(target);
        } catch (IllegalArgumentException e) {
            throw new GluonJRuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new GluonJRuntimeException(e);
        }
    }

    /**
     * Removes <tt>aspect</tt> object from the anonymous field of
     * <tt>target</tt> object. <tt>clazz</tt> object specifies the type of
     * anonymous field.
     */
    public static boolean remove(Object target, Object aspect, Class clazz) {
        Field f = getAF(target.getClass(), clazz);
        try {
            if (f.get(target).equals(aspect)) {
                f.set(target, null);
                return true;
            } else {
                return false;
            }
        } catch (IllegalArgumentException e) {
            throw new GluonJRuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new GluonJRuntimeException(e);
        }

    }

    private static Field getAF(Class class1, Class clazz) {
        String prefix = getAFPreFix(clazz.getName(), class1.getName());
        for (int i = 0; i < class1.getDeclaredFields().length; i++) {
            Field f = class1.getDeclaredFields()[i];
            if (f.getName().startsWith(prefix)) {
                f.setAccessible(true);
                return f;
            }
        }
        for (int i = 0; i < class1.getFields().length; i++) {
            Field f = class1.getFields()[i];
            if (f.getName().startsWith(prefix)) {
                f.setAccessible(true);
                return f;
            }
        }
        throw new GluonJRuntimeException(new NoSuchFieldException(class1
                .getName()
                + " does not have a field which name starts with " + prefix));
    }

    public static String getAFPreFix(String type, String targetType) {

        int typeId = type.hashCode();
        int declaringId = targetType.trim().hashCode();

        typeId = Math.abs(typeId);
        declaringId = Math.abs(declaringId);
        return "aop$gluonj$aspect$" + typeId + "$" + declaringId;

    }
    
    public static boolean isPerCflow(String fName) {
        return fName.startsWith(PERCFLOW);
    }
    public static boolean isPerObjects(String fName) {
        return fName.startsWith(PEROBJECTS);
    }
    
    public static String PERCFLOW = "$percflow";
    
    public static String PEROBJECTS = "$perobjects";
}