/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.ui;

import gluonj.NotFoundException;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.regex.Pattern;

class FileFinder {


    private String[] allFiles(File base) {
        if (base.isDirectory()) {
            File[] children = base.listFiles();
            List pathList = new LinkedList();
            for (int i = 0; i < children.length; i++) {
                pathList.addAll(Arrays.asList(allFiles(children[i])));
            }
            return (String[]) pathList.toArray(new String[pathList.size()]);
        } else {
            return new String[] { base.getAbsolutePath() };
        }
    }

    /*
     * classPattern is like usr/test/Test, usr/test/Main$1, usr/test/Foo$Bar
     * 
     * jarPath is like lib\impl.jar, c:\lib\impl.jar (Windows)
     */
    String[] resolveClassPatInJar(String classPattern, String jarPath)
            throws NotFoundException {

        List l = new LinkedList();
        JarFile jar;
        try {
            jar = new JarFile(new File(jarPath));
        } catch (IOException e) {
            throw new NotFoundException("Jar file \"" + jarPath
                    + "\" is not found.", e);
        }

        for (Enumeration entries = jar.entries(); entries.hasMoreElements();) {
            JarEntry entry = (JarEntry) entries.nextElement();
            String name = entry.getName();
            if (name.endsWith(".class")) {
                String className = name.substring(0, name.length() - 6);

                if (match(classPattern, className)) {
                    l.add(className.replace('/', '.'));
                }
            }
        }
        return (String[]) l.toArray(new String[l.size()]);

    }

    /*
     * format of pattern is
     * 
     * "*" means more than zero characters except SEPARATOR.
     * 
     * "**" means more than zero characters. this wildcard must exist between
     * two SEPARATORs. TestExec: "usr/test/** /Main"
     * 
     * 
     * a targetName is like usr/test/Test, usr/test/Main$1, usr/test/Foo$Bar
     * 
     * 
     * 
     * 
     * 
     * "*" -> "[^/]*" "/**" + "/ " -> "/.*" + "/ | /"
     * 
     * (algorithm)
     * 
     * "*" -> "[^/]*" "/[^/]*[^/]*" + "/ " -> "(/.*" + "/ | /)"
     * 
     * "." -> "\." "(" -> "\(" ")" -> "\)" "^" -> "\^" "+" -> "\+" "{" -> "\{"
     * "}" -> "\}"
     * 
     *  
     */

    boolean match(String pattern, String targetName) {

        pattern = pattern.replaceAll("\\*", "[^/]*");
        pattern = pattern.replaceAll("\\/\\[\\^\\/\\]\\*\\[\\^\\/\\]\\*\\/",
                "(/.*/|/)");

        return Pattern.matches(pattern, targetName);
    }

    boolean endsWith(String pattern, String targetName) {

        pattern = pattern.replaceAll("\\*", "[^/]*");
        pattern = pattern.replaceAll("\\/\\[\\^\\/\\]\\*\\[\\^\\/\\]\\*\\/",
                "(/.*/|/)");

        return Pattern.matches(".*" + pattern, targetName);
    }

    /*
     * classPattern is like "usr/test/Test", "usr/test/Main$1", "usr/test/Foo$Bar"
     * 
     * dirPath is like "lib", "c:\lib" (Windows)
     * 
     * returns class name like "java.util.List"  
     */
    String[] resolveClassPatInDir(String classPattern, String dirPath) {

        
        String base = dirPath.replace('\\', '/')  +"/";
        List l = new LinkedList();
        String[] filePath = allFiles(new File(dirPath));

        for (int i = 0; i < filePath.length; i++) {
            String name = filePath[i];
            if (name.endsWith(".class")) {
                name = name.substring(0, name.length() - 6);
                name = name.replace('\\', '/');
                name = name.replaceFirst(base, "");
                if (endsWith(classPattern, name)) {
                    l.add(name.replace('/', '.'));
                }
            }
        }
        
        return (String[]) l.toArray(new String[l.size()]);
    }
}