/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.ui;

import gluonj.ast.Aspect;
import gluonj.parser.AspectParser;
import gluonj.parser.ParseException;
import gluonj.weaver.AOPWeaver;
import gluonj.weaver.AOPWeaverFactory;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;

import javassist.CannotCompileException;
import javassist.ClassPath;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;

public class StaticWeaver {

    private static OptionTable.Factory init() {
        OptionTable.Factory factory = null;
        factory = new OptionTable.Factory();

        factory.setOptionFormat(opt_src, 1);
        factory.setOptionFormat(opt_incl, 1);
        factory.setOptionFormat(opt_cp, 1);
        factory.setOptionFormat(opt_injar, 1);
        factory.setOptionFormat(opt_dest, 1);
        factory.setOptionFormat(opt_outjar, 1);
        return factory;
    }

    public static String opt_src = "classdir";

    public static String opt_incl = "includes";

    public static String opt_cp = "classpath";

    public static String opt_injar = "injar";

    public static String opt_dest = "destdir";

    public static String opt_outjar = "outjar";

    public static void weave(String[] args) throws ParseException, gluonj.weaver.CannotCompileException, gluonj.NotFoundException, NotFoundException, IOException {

        if(args.length == 0) {
            printHelp();
            return;
        }
        
        OptionTable.Factory factory = init();

        /*
         * tokenize the option
         */
        OptionTable table = factory.make(args);

        /*
         * tokenize the filelist
         */
        String[] arguments = table.getArgument();

        StaticWeaver weaver = new StaticWeaver();

        weaver.setSrcDir(fileArrayOf(table, opt_src));
        weaver.setIncludes(stringOf(table, opt_incl));
        weaver.setClassPath(fileArrayOf(table, opt_cp));
        weaver.setInJar(fileArrayOf(table, opt_injar));
        weaver.setDestDir(fileOf(table, opt_dest));
        weaver.setOutJar(fileOf(table, opt_outjar));
        weaver.setGluePath(fileArrayOf(arguments));

        weaver.weave();
    }

    private static void printHelp() {
        StringBuffer buf = new StringBuffer();
        buf.append("usage: ");
        buf.append("[-classdir <dirs>] ");
        buf.append("[-injar <jarfiles>] ");
        buf.append("[-classpath <pathes>] ");
        buf.append("[-includes <patterns>] ");
        buf.append("[-dest <path>] ");
        buf.append("[-outjar <jarfile>] ");
        buf.append("<glue1> <glue2> ...");
        System.out.println(buf.toString());
    }

    private static File fileOf(OptionTable table, String dest) {
        if (table.contains(dest)) {
            return new File(table.getValueOf(dest)[0]);
        } else {
            return null;
        }
    }

    private static File[] fileArrayOf(String[] arguments) {
        File[] fs;
        fs = new File[arguments.length];
        for (int i = 0; i < arguments.length; i++) {
            fs[i] = new File(arguments[i]);
        }
        return fs;
    }

    private static File[] fileArrayOf(OptionTable table, String src) {
        if (table.contains(src)) {
            String d = table.getValueOf(src)[0];
            String[] ds = d.split(File.pathSeparator);
            File[] fs;
            fs = new File[ds.length];
            for (int i = 0; i < ds.length; i++) {
                fs[i] = new File(ds[i]);
            }
            return fs;
        } else {
            return null;
        }
    }

    private static String[] stringOf(OptionTable table, String src) {
        if (table.contains(src)) {
            String d = table.getValueOf(src)[0];
            String[] ds = d.split(File.pathSeparator);
            return ds;
        } else {
            return null;
        }
    }

    private static String[] stringOf(File[] files) {
        String[] str = new String[files.length];
        for (int i = 0; i < str.length; i++) {
            str[i] = files[i].getAbsolutePath();
        }
        return str;
    }

    public StaticWeaver() {
    }

    private File[] basedir;

    private File[] injar;

    private String[] includes;

    private File[] classpath;

    private File destdir;

    private File outjar;

    private File[] glue;

    public void setSrcDir(File[] files) {
        if (files == null) {
            files = new File[0];
        }
        this.basedir = files;
    }

    public void setInJar(File[] files) {
        if (files == null) {
            files = new File[0];
        }
        injar = files;
    }

    public void setIncludes(String[] files) {
        if (files == null) {
            files = new String[0];
        }
        includes = files;
    }

    public void setClassPath(File[] files) {
        if (files == null) {
            files = new File[0];
        }
        classpath = files;
    }

    public void setDestDir(File dir) {
        destdir = dir;
    }

    public void setOutJar(File outjar) {
        this.outjar = outjar;
    }

    public void setGluePath(File[] files) {
        if (files == null) {
            files = new File[0];
        }
        this.glue = files;
    }

    public void weave() throws ParseException,
            gluonj.weaver.CannotCompileException, gluonj.NotFoundException,
            NotFoundException, IOException {

        ClassPool pool = new ClassPool();
        pool.appendSystemPath();

        /* defines where to output */
        System.out.print("destination: ");
        ClassWritingHandler handler;

        if (outjar != null) {
            outjar.createNewFile();
            handler = new JarWriter(outjar);
            System.out.println(outjar.getAbsolutePath());
        } else if (destdir != null) {
            destdir.mkdir();
            handler = new FileWriter(destdir);
            System.out.println(destdir.getAbsolutePath());
        } else {
            handler = new FileWriter(new File("."));
            System.out.println(".");
        }

        /* gets the glue path */
        String[] gluePath = stringOf(glue);

        /* searches class names from srcdir */
        String[] className = searchClass(basedir, includes);

        /* searches class names from jar file */
        String[] classPathInJar = searchClassJar(injar, includes);
        className = append(className, classPathInJar);

        /* appends class path and jar file to ClassPool */
        System.out.print("classpath: ");
        for (int i = 0; i < basedir.length; i++) {
            System.out.print(basedir[i].getAbsolutePath() + ", ");
            pool.appendClassPath(basedir[i].getAbsolutePath());
        }
        for (int i = 0; i < injar.length; i++) {
            System.out.print(injar[i].getAbsolutePath() + ", ");
            pool.appendClassPath(injar[i].getAbsolutePath());
        }
        for (int i = 0; i < classpath.length; i++) {
            System.out.print(classpath[i].getAbsolutePath() + ", ");
            pool.appendClassPath(classpath[i].getAbsolutePath());
        }
        System.out.println("");

        System.out.print(gluePath.length + " glue files: ");
        for (int i = 0; i < gluePath.length; i++) {
            System.out.print(gluePath[i] + ", ");
        }
        System.out.println("");
        System.out.print(className.length + " class files: ");
        for (int i = 0; i < className.length; i++) {
            System.out.print(className[i] + ", ");
        }
        System.out.println("");
        weave(gluePath, className, pool, handler);
    }

    private String[] append(String[] str1, String[] str2) {
        String[] appended = new String[str1.length + str2.length];
        for (int i = 0; i < str1.length; i++) {
            appended[i] = str1[i];
        }
        for (int i = str1.length; i < str1.length + str2.length; i++) {
            appended[i] = str2[i - str1.length];
        }
        return appended;
    }

    /*
     * aspectName: FilePath (c:\test\test.xml), targetclassName: class
     * (test.MyTest)
     *  
     */
    public static void weave(String[] aspectName, String[] targetClassName,
            ClassPool pool, ClassWritingHandler handler) throws ParseException,
            gluonj.weaver.CannotCompileException {
        try {

            Aspect[] aspects = new Aspect[aspectName.length];
            for (int i = 0; i < aspectName.length; i++) {
                aspects[i] = AspectParser.factory().parse(aspectName[i]);
            }
            AOPWeaver weaver = new AOPWeaverFactory().make(aspects);
            Set createdClassSet = new HashSet();
            weaver.setCreatedClassSet(createdClassSet);
            weaver.weaveIntertypeDecl(pool);
            for (int i = 0; i < targetClassName.length; i++) {
                CtClass cc = pool.get(targetClassName[i]);
                weaver.weaveAdvice(cc, pool);
                handler.onWrite(cc);
            }
            for (Iterator iter = createdClassSet.iterator(); iter.hasNext();) {
                handler.onWrite((CtClass) iter.next());
            }
            handler.finish();

        } catch (NotFoundException e) {
            throw new gluonj.weaver.CannotCompileException(
                    "caused when writing bytecode.", e);
        } catch (CannotCompileException e) {
            throw new gluonj.weaver.CannotCompileException(
                    "caused when writing bytecode.", e);
        } catch (IOException e) {
            throw new gluonj.weaver.CannotCompileException(
                    "caused when writing bytecode.", e);
        }
    }

    public interface ClassWritingHandler {
        public void onWrite(CtClass cc) throws NotFoundException, IOException,
                CannotCompileException;

        public void finish() throws IOException;
    }

    public static void weave(String[] aspectName, String[] targetClassName,
            final String dest) throws ParseException,
            gluonj.weaver.CannotCompileException {

        ClassWritingHandler writer = new ClassWritingHandler() {
            public void onWrite(CtClass cc) throws NotFoundException,
                    IOException, CannotCompileException {

                cc.writeFile(dest);
            }

            public void finish() throws IOException {
            }
        };

        weave(aspectName, targetClassName, ClassPool.getDefault(), writer);
    }

    /*
     * includes: "java/util/*", "**", "test/**\backslash;Test*" etc. return:
     * "java.util.List","java.lang.Object" etc.
     */

    private String[] searchClass(File[] basedir, String[] includes) {

        Set classPat = new HashSet();
        for (int i = 0; i < basedir.length; i++) {
            for (int j = 0; j < includes.length; j++) {
                classPat.addAll(Arrays.asList(new FileFinder()
                        .resolveClassPatInDir(includes[j], basedir[i]
                                .getAbsolutePath())));
            }
        }
        return (String[]) classPat.toArray(new String[classPat.size()]);

    }

    private String[] searchClassJar(File[] basedir, String[] includes)
            throws gluonj.NotFoundException {

        List classPat = new LinkedList();
        for (int i = 0; i < basedir.length; i++) {
            for (int j = 0; j < includes.length; j++) {
                classPat.addAll(Arrays.asList(new FileFinder()
                        .resolveClassPatInJar(includes[j], basedir[i]
                                .getAbsolutePath())));
            }
        }
        return (String[]) classPat.toArray(new String[classPat.size()]);

    }

    static class FileWriter implements ClassWritingHandler {
        String dest;

        FileWriter(File f) {
            dest = f.getAbsolutePath();
        }

        public void onWrite(CtClass cc) throws NotFoundException, IOException,
                CannotCompileException {
            cc.writeFile(dest);
        }

        public void finish() throws IOException {
        }
    }

    static class JarWriter implements ClassWritingHandler {
        JarOutputStream jarout;

        JarWriter(File jarPath) throws IOException {
            jarout = new JarOutputStream(new FileOutputStream(jarPath));
        }

        public void onWrite(CtClass cc) throws NotFoundException, IOException,
                CannotCompileException {

            String filename = cc.getName().replace('.', '/') + ".class";
            jarout.putNextEntry(new JarEntry(filename));
            byte[] b = cc.toBytecode();
            jarout.write(b);
            jarout.flush();
            jarout.closeEntry();
        }

        public void finish() throws IOException {
            jarout.finish();
            jarout.close();
        }
    }
}