/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.weaver;

import gluonj.ast.Advice;
import gluonj.ast.AdviceInfo;
import gluonj.ast.AdviceType;
import gluonj.ast.intertype.IntertypeDecl;
import gluonj.ast.java.ASTBlock;
import gluonj.ast.java.ASTExpr;
import gluonj.ast.java.ASTFormalParameter;
import gluonj.ast.joinpoint.JPConstructor;
import gluonj.ast.joinpoint.JPFactory;
import gluonj.ast.joinpoint.JPFieldAccess;
import gluonj.ast.joinpoint.JPMethod;
import gluonj.ast.joinpoint.JPMethodCall;
import gluonj.ast.joinpoint.JPNewExpr;
import gluonj.ast.joinpoint.JoinPoint;
import gluonj.ast.pattern.DotTypeNamePattern;
import gluonj.ast.pattern.IdPattern;
import gluonj.ast.pattern.IdentifierTypeNamePattern;
import gluonj.ast.pattern.SingleTypeNamePattern;
import gluonj.ast.pcs.PcsConstructor;
import gluonj.ast.pcs.PcsModifier;
import gluonj.ast.pcs.PcsParameter;
import gluonj.ast.pcs.PcsParameterDotDot;
import gluonj.ast.pcs.PcsType;
import gluonj.ast.pcs.PointcutResidue;
import gluonj.ast.pointcut.ASTPCShadow;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.weaver.FieldWeavingVisitor.DuplicationInfo;

import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.FieldAccess;
import javassist.expr.MethodCall;
import javassist.expr.NewExpr;

public class AOPWeaver {

    static private final String PROCEED = "$_ = $proceed($$);";

    private List intertypeDeclList;

    private List adviceList;

    public AOPWeaver() {
        init();
    }

    private void init() {
        intertypeDeclList = new LinkedList();
        adviceList = new LinkedList();
    }

    /*
     * pointcut DNFA1AND Ɋ܂܂̂ ASTShadow  ASTResidue ̂݁B
     * ASTShadow.getPointcutShadow() ŕԂCX^X͑Sē
     */
    public void addAdvice(Advice adv) {
        adviceList.add(adv);
    }

    public void addIntertypeDecl(IntertypeDecl decl) {
        intertypeDeclList.add(decl);

    }

    public void weaveAdvice(CtClass cc, ClassPool pool)
            throws CannotCompileException {

        CtMethod[] cm = cc.getDeclaredMethods();
        CtConstructor[] ccn = cc.getDeclaredConstructors();
        CtConstructor init = cc.getClassInitializer();

        if (!createdCtClass.contains(cc)) {
            for (int i = 0; i < cm.length; i++) {
                if (!Modifier.isAbstract(cm[i].getModifiers())) {
                    weaveToBehaviorBody(cm[i], pool);
                }
            }
            for (int i = 0; i < ccn.length; i++) {
                weaveToBehaviorBody(ccn[i], pool);
            }
            if (init != null) {
                weaveToBehaviorBody(init, pool);
            }
        }
        for (int i = 0; i < cm.length; i++) {
            if (!Modifier.isAbstract(cm[i].getModifiers())) {
                weaveToMethod(cm[i], pool);
            }
        }
        for (int i = 0; i < ccn.length; i++) {
            weaveToConstructor(ccn[i], pool);
        }
        if (init != null) {
            weaveToConstructor(init, pool);
        }
    }

    private void weaveToBehaviorBody(CtBehavior cb, final ClassPool pool)
            throws CannotCompileException {

        cb.instrument(new ExprEditor() {
            public void edit(final FieldAccess expr)
                    throws CannotCompileException {

                JoinPoint jp = JPFactory.makeJP(expr);
                WeavingProcess p = new WeavingProcess() {
                    public void atAround(StringBuffer codeBefore,
                            StringBuffer codeAfter)
                            throws CannotCompileException {
                        try {
                            CtField cf = expr.getField();
                            CtClass closure = new ClosureMaker(pool)
                                    .attachClosure(expr.where()
                                            .getDeclaringClass(), cf, expr
                                            .isWriter());
                            String push = makeClosurePushCode(closure);
                            String pop = makeClosurePopCode();
                            codeBefore.append(push);
                            codeAfter.append(pop);
                        } catch (NotFoundException e) {
                            throw new CannotCompileException(e);
                        }
                    }

                    public void atWeave(String codeBefore, String codeAfter,
                            String codeAround) throws CannotCompileException {

                        if (codeAround == "") {
                            codeAround = PROCEED;
                        }
                        expr.replace("{" + codeBefore + codeAround + codeAfter
                                + "}");
                    }
                };
                weave(jp, p, pool);
            }

            public void edit(final MethodCall expr)
                    throws CannotCompileException {

                JoinPoint jp = JPFactory.makeJP(expr);
                WeavingProcess p = new WeavingProcess() {
                    public void atAround(StringBuffer codeBefore,
                            StringBuffer codeAfter)
                            throws CannotCompileException {

                        try {
                            CtMethod callee = expr.getMethod();
                            CtClass closure = new ClosureMaker(pool)
                                    .attachClosure(expr.where()
                                            .getDeclaringClass(), callee);
                            String push = makeClosurePushCode(closure);
                            String pop = makeClosurePopCode();
                            codeBefore.append(push);
                            codeAfter.append(pop);
                        } catch (NotFoundException e) {
                            throw new CannotCompileException(e);
                        }
                    }

                    public void atWeave(String codeBefore, String codeAfter,
                            String codeAround) throws CannotCompileException {

                        if (codeAround == "") {
                            codeAround = PROCEED;
                        }
                        expr.replace("{" + codeBefore + codeAround + codeAfter
                                + "}");
                    }
                };
                weave(jp, p, pool);
            }

            public void edit(final NewExpr expr) throws CannotCompileException {
                JoinPoint jp = JPFactory.makeJP(expr);
                WeavingProcess p = new WeavingProcess() {
                    public void atAround(StringBuffer codeBefore,
                            StringBuffer codeAfter)
                            throws CannotCompileException {

                    }

                    public void atWeave(String codeBefore, String codeAfter,
                            String codeAround) throws CannotCompileException {

                        if (codeAround == "") {
                            codeAround = PROCEED;
                        }
                        expr.replace("{" + codeBefore + codeAround + codeAfter
                                + "}");
                    }
                };
                weave(jp, p, pool);

            }
        });
    }

    private void weaveToConstructor(final CtConstructor ccn, ClassPool pool)
            throws CannotCompileException {

        JoinPoint jp = JPFactory.makeJP(ccn);
        WeavingProcess p = new WeavingProcess() {
            public void atAround(StringBuffer codeBefore, StringBuffer codeAfter)
                    throws CannotCompileException {

                throw new UnSupportedException(
                        "around advice to constructor-exec is not supported.");
            }

            public void atWeave(String codeBefore, String codeAfter,
                    String codeAround) throws CannotCompileException {

                if (codeAround.equals("") || codeAround == null) {
                    if (!(codeBefore.equals("") || codeBefore == null)) {
                        ccn.insertBefore(codeBefore);
                    }
                    if (!(codeAfter.equals("") || codeAfter == null)) {
                        ccn.insertAfter(codeAfter);
                    }
                } else {
                    ccn
                            .setBody("{" + codeBefore + codeAround + codeAfter
                                    + "}");
                }
            }
        };
        weave(jp, p, pool);
    }

    private void weaveToMethod(final CtMethod cm, final ClassPool pool)
            throws CannotCompileException {

        JoinPoint jp = JPFactory.makeJP(cm);
        WeavingProcess p = new WeavingProcess() {
            public void atAround(StringBuffer codeBefore, StringBuffer codeAfter)
                    throws CannotCompileException {

                CtMethod cmcopy = CtNewMethod.copy(cm, cm.getName() + "$copy",
                        cm.getDeclaringClass(), null);
                int mod = cmcopy.getModifiers();
                cmcopy.setModifiers(javassist.Modifier.setPublic(mod));
                cm.getDeclaringClass().addMethod(cmcopy);

                CtClass closure = new ClosureMaker(pool).attachClosure(cmcopy
                        .getDeclaringClass(), cmcopy);
                String push = makeClosurePushCode(closure);
                String pop = makeClosurePopCode();
                codeBefore.append(push);
                codeAfter.append(pop);
            }

            public void atWeave(String codeBefore, String codeAfter,
                    String codeAround) throws CannotCompileException {

                if (codeAround.equals("") || codeAround == null) {
                    if (!(codeBefore.equals("") || codeBefore == null)) {
                        cm.insertBefore(codeBefore);
                    }
                    if (!(codeAfter.equals("") || codeAfter == null)) {
                        cm.insertAfter(codeAfter);
                    }
                } else {
                    cm.setBody("{" + codeBefore + codeAround + codeAfter + "}");
                }
            }
        };
        weave(jp, p, pool);
    }

    private void weave(JoinPoint jp, WeavingProcess p, ClassPool pool)
            throws CannotCompileException {

        List matchedAdvices = new LinkedList();
        StringBuffer codeBefore = new StringBuffer("");
        StringBuffer codeAfter = new StringBuffer("");
        String codeAround = "";
        boolean match = false;

        try {
            for (Iterator iter = adviceList.iterator(); iter.hasNext();) {
                Advice advice = (Advice) iter.next();

                AdviceInfo info = advice.match(jp);

                if (info != null) {
                    matchedAdvices.add(advice);
                    for (int i = 0; i < advice.getPackage().length; i++) {
                        pool.importPackage(advice.getPackage()[i].getPath());
                    }
                    AdviceType type = info.getType();
                    if (type == AdviceType.AROUND) {
                        p.atAround(codeBefore, codeAfter);
                    }
                    match = true;
                    String body = makeTJPCode(jp, info.getBody())
                            + info.getBody();
                    if (type == AdviceType.BEFORE) {
                        codeBefore.insert(0, body);
                    } else if (type == AdviceType.AFTER) {
                        codeAfter.append(body);
                    } else if (type == AdviceType.AROUND) {
                        codeAround = body;
                    }
                }
            }

            if (match) {
                p.atWeave(codeBefore.toString(), codeAfter.toString(),
                        codeAround);
            }
        } catch (CannotCompileException e) {
            StringBuffer buf = new StringBuffer();
            for (Iterator iter = matchedAdvices.iterator(); iter.hasNext();) {
                Advice adv = (Advice) iter.next();
                buf.append("{" + adv.getBody().toString() + "},");
            }
            throw new CannotCompileExceptionInner(e, buf.toString());
        } catch (NotFoundException e) {
            throw new CannotCompileException(e);
        } finally {
            pool.clearImportedPackages();
        }
    }

    private Set createdCtClass = new HashSet();

    public void setCreatedClassSet(Set s) {
        createdCtClass = s;
    }

    private String makeClosurePushCode(CtClass closure) {

        createdCtClass.add(closure);

        StringBuffer buf = new StringBuffer();
        buf.append("{");

        buf.append("gluonj.runtime.Closure c = new " + closure.getName()
                + "();");

        /*
         * buf .append("gluonj.runtime.Closure c = new
         * gluonj.runtime.ReflectionClosure($class, \"" + cm.getName() +
         * "\",$sig);\n");
         */
        buf.append("c.setThisObject($0);\n");
        buf.append("c.setArguments($args);\n");
        buf.append("gluonj.runtime.ClosureStack.push(c);\n");
        buf.append("\n}");
        return buf.toString();
    }

    private String makeClosurePopCode() {
        return "{ gluonj.runtime.ClosureStack.pop(); }";
    }

    private static interface WeavingProcess {
        void atAround(StringBuffer codeBefore, StringBuffer codeAfter)
                throws CannotCompileException;

        void atWeave(String codeBefore, String codeAfter, String codeAround)
                throws CannotCompileException;
    }

    public void weaveIntertypeDecl(CtClass cc, ClassPool pool)
            throws gluonj.weaver.CannotCompileException {
        IntertypeDeclWeavingVisitor v;

        try {
            v = new SkeletonWeavingVisitor(pool);
            applyVisitor(cc, v);
            FieldWeavingVisitor fv = new FieldWeavingVisitor(pool);
            applyVisitor(pool, fv);
            addDiAdvice(fv.getDuplication());
            v = new BodyWeavingVisitor(pool);
            applyVisitor(cc, v);
        } catch (CannotCompileException e) {
            throw new gluonj.weaver.CannotCompileException(e);
        }
    }

    public void weaveIntertypeDecl(ClassPool pool)
            throws gluonj.weaver.CannotCompileException {

        IntertypeDeclWeavingVisitor v;

        try {
            v = new SkeletonWeavingVisitor(pool);
            applyVisitor(pool, v);
            FieldWeavingVisitor fv = new FieldWeavingVisitor(pool);
            applyVisitor(pool, fv);
            addDiAdvice(fv.getDuplication());
            v = new BodyWeavingVisitor(pool);
            applyVisitor(pool, v);
        } catch (CannotCompileException e) {
            throw new gluonj.weaver.CannotCompileException(e);
        }
    }

    private void addDiAdvice(DuplicationInfo[] dupInfo)
            throws CannotCompileException {

        for (int i = 0; i < dupInfo.length; i++) {

            CtField cf = dupInfo[i].getDupField();
            int flagId = Math.abs(dupInfo[i].hashCode());
            Advice[] adv = makeDiAdvice(cf, dupInfo[i].getInitializer(), flagId);
            adviceList.addAll(Arrays.asList(adv));
        }
    }

    private Advice[] makeDiAdvice(CtField dupField, ASTExpr initializer,
            int flagId) throws CannotCompileException {

        if (Modifier.isStatic(dupField.getModifiers())) {
            CtConstructor ccn = dupField.getDeclaringClass()
                    .makeClassInitializer();
            String code = makeInjectionCode(dupField, initializer);
            ccn.insertAfter(code);
            return new Advice[0];
        } else {
            CtClass cc = dupField.getDeclaringClass();
            PcsConstructor pcn = makePcsConstructor(cc);

            ASTPointcutExpr e = new ASTPCShadow(pcn, PointcutResidue.TRUE);

            Advice before = new Advice(new ASTFormalParameter[0],
                    AdviceType.BEFORE, e, new ASTBlock(makeBodyBefore(dupField,
                            initializer, flagId)));

            Advice after = new Advice(new ASTFormalParameter[0],
                    AdviceType.AFTER, e, new ASTBlock(makeBodyAfter(dupField,
                            initializer, flagId)));

            return new Advice[] {
                    before,
                    after };
        }
    }

    private String makeBodyBefore(CtField dupField, ASTExpr initializer,
            int flagId) {
        StringBuffer buf = new StringBuffer();
        buf.append("gluonj.runtime.CflowCounter.enter(" + flagId + ");");
        return buf.toString();
    }

    private String makeBodyAfter(CtField dupField, ASTExpr initializer,
            int flagId) {
        StringBuffer buf = new StringBuffer();
        buf.append("gluonj.runtime.CflowCounter.getout(" + flagId + ");");
        buf.append("if(!gluonj.runtime.CflowCounter.isIn(" + flagId + ")) {");
        buf.append(makeInjectionCode(dupField, initializer));
        buf.append("}");
        return buf.toString();
    }

    private String makeInjectionCode(CtField dupField, ASTExpr initializer) {
        String name = dupField.getName();
        return "{" + name + " = " + initializer.getExpr() + ";}";
    }

    private PcsConstructor makePcsConstructor(CtClass cc) {

        String[] ns = cc.getName().split("\\.");
        SingleTypeNamePattern namePat = new IdentifierTypeNamePattern(
                new IdPattern(ns[0]));
        for (int i = 1; i < ns.length; i++) {
            namePat = new DotTypeNamePattern(namePat,
                    new IdentifierTypeNamePattern(new IdPattern(ns[i])));
        }

        return new PcsConstructor(new PcsModifier(), new PcsType(
                new PcsModifier(), namePat),
                new PcsParameter[] { new PcsParameterDotDot() }, new PcsType[0]);
    }

    private void applyVisitor(CtClass cc, IntertypeDeclWeavingVisitor visitor)
            throws CannotCompileException {

        for (Iterator iter = intertypeDeclList.iterator(); iter.hasNext();) {

            IntertypeDecl decl = (IntertypeDecl) iter.next();
            String className = decl.getClassPattern().getSoleClassName();
            try {
                applyVisitor(decl, cc, visitor);
            } catch (RuntimeException e) {
                if (e.getCause() instanceof CannotCompileException) {
                    throw (CannotCompileException) e.getCause();
                }
            }
        }
    }

    private void applyVisitor(ClassPool pool,
            IntertypeDeclWeavingVisitor visitor) throws CannotCompileException {

        for (Iterator iter = intertypeDeclList.iterator(); iter.hasNext();) {

            IntertypeDecl decl = (IntertypeDecl) iter.next();
            String className = decl.getClassPattern().getSoleClassName();
            CtClass cc;
            try {
                cc = pool.get(className);
            } catch (NotFoundException e) {
                continue;
            }
            applyVisitor(decl, cc, visitor);
        }
    }

    private void applyVisitor(IntertypeDecl decl, CtClass cc,
            IntertypeDeclWeavingVisitor visitor) throws CannotCompileException {

        try {
            if (decl.getClassPattern().match(cc)) {
                visitor.setTargetClass(cc);
                for (int i = 0; i < decl.getImport().length; i++) {
                    cc.getClassPool().importPackage(decl.getImport()[i].getPath());
                }
                decl.getMember().accept(visitor);
            }
        } catch (RuntimeException e) {
            if (e.getCause() instanceof CannotCompileException) {
                throw (CannotCompileException) e.getCause();
            } else {
                throw new CannotCompileException(e.getCause());
            }
        } finally {
            cc.getClassPool().clearImportedPackages();
        }
    }

    static public class CannotCompileExceptionInner extends

    CannotCompileException {

        public CannotCompileExceptionInner(CannotCompileException e, String body) {
            super(e);
            this.body = body;
        }

        private String body;

        public String getBody() {
            return body;
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("advices: " + adviceList.size() + "\n");
        for (Iterator iter = adviceList.iterator(); iter.hasNext();) {
            buf.append(iter.next() + "\n");
        }
        buf.append("ITDs: " + intertypeDeclList.size() + "\n");
        for (Iterator iter = intertypeDeclList.iterator(); iter.hasNext();) {
            buf.append(iter.next() + "\n");
        }
        return buf.toString();
    }

    private String makeTJPCode(JoinPoint jp, String body)
            throws NotFoundException {

        StringBuffer buf = new StringBuffer();

        if (body.indexOf("$thisCflow") != -1) {
            buf
                    .append("gluonj.runtime.Cflow $thisCflow = new gluonj.runtime.Cflow();");
        }
        if (jp.getJPType() == JoinPoint.METHOD) {
            CtMethod cm = ((JPMethod) jp).getCtMethod();
            buf.append("String $name = \"" + getSignature(cm) + "\";");

        } else if (jp.getJPType() == JoinPoint.CONSTRUCTOR) {

            CtConstructor ccst = ((JPConstructor) jp).getCtConstructor();
            buf.append("String $name = \"" + getSignature(ccst) + "\";");

        } else if (jp.getJPType() == JoinPoint.METHOD_CALL) {

            MethodCall expr = ((JPMethodCall) jp).getMethodCall();
            buf.append("String $name = \"" + getSignature(expr.getMethod())
                    + "\";");
            buf.append("String $where = \"" + getSignature(expr.where())
                    + "\";");

        } else if (jp.getJPType() == JoinPoint.FIELD_ACCESS) {
            FieldAccess expr = ((JPFieldAccess) jp).getFieldAccess();
            buf.append("String $name = \"" + getSignature(expr.getField())
                    + "\";");
            buf.append("String $where = \"" + getSignature(expr.where())
                    + "\";");

        } else if (jp.getJPType() == JoinPoint.NEW_EXPR) {
            NewExpr expr = ((JPNewExpr) jp).getNewExpr();
            buf.append("String $name = \""
                    + getSignature(expr.getConstructor()) + "\";");
            buf.append("String $where = \"" + getSignature(expr.where())
                    + "\";");
        } else {
            throw new RuntimeException("unsupported jp." + jp);
        }
        return buf.toString();
    }

    private String getNames(CtClass[] params) {

        StringBuffer buf = new StringBuffer();

        if (params.length > 0) {
            buf.append(params[0].getName());
            for (int i = 1; i < params.length; i++) {
                buf.append(',');
                buf.append(params[i].getName());
            }
        }
        return buf.toString();
    }

    private String getSignature(CtBehavior cb) throws NotFoundException {

        String returnTypeName = "";
        if (cb instanceof CtMethod) {
            returnTypeName = ((CtMethod) cb).getReturnType().getName() + " ";
        }
        String className = cb.getDeclaringClass().getName();
        String name;
        if (cb instanceof CtMethod) {
            name = cb.getName();
        } else if (cb instanceof CtConstructor) {
            name = "new";
        } else {
            throw new RuntimeException();
        }

        String paramNames = getNames(cb.getParameterTypes());
        String expNames = getNames(cb.getExceptionTypes());

        StringBuffer buf = new StringBuffer();
        buf.append(returnTypeName + className + "." + name + "(" + paramNames
                + ")");
        if (!expNames.equals("")) {
            buf.append(" throws " + expNames);
        }
        return buf.toString();
    }

    private String getSignature(CtField cm) throws NotFoundException {

        String typeName = cm.getType().getName();
        String className = cm.getDeclaringClass().getName();
        String methodName = cm.getName();

        StringBuffer buf = new StringBuffer();
        buf.append(typeName + " " + className + "." + methodName);
        return buf.toString();
    }
}