/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.weaver;

import java.util.HashMap;
import java.util.Map;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;

/**
 * make subclass of gluonj.runtime.Closure.
 */
public class ClosureMaker {

    private ClassPool pool;

    public ClosureMaker(ClassPool p) {
        pool = p;
    }

    /**
     * Attaches a closure class to target class. New closure class calls method
     * represented by callee. the closure class is a subclass of
     * gluonj.runtime.Closure.
     * 
     * @param target :
     *            target class which closure class is attached to.
     * @param callee :
     *            called in closure.
     * 
     * @return attached closure class.
     */
    public CtClass attachClosure(CtClass target, CtMethod callee) {
        try {
            String proceedBody = makeProceedBody(callee);
            String closureName = makeClosureName(target, callee);
            return attachClosure(target, closureName, proceedBody);
        } catch (NotFoundException e) {
            throw new InternalError(e);
        }
    }

    public CtClass attachClosure(CtClass target, CtField field, boolean isSetter) {
        try {
            String proceedBody = makeProceedBody(field, isSetter);
            String closureName = makeClosureName(target, field, isSetter);
            return attachClosure(target, closureName, proceedBody);
        } catch (NotFoundException e) {
            throw new InternalError(e);
        }
    }

    private CtClass attachClosure(CtClass target, String closureName,
            String proceedBody) {
        try {
            CtClass ccClosure = target.makeNestedClass(closureName, true);
            ccClosure.setSuperclass(pool.get("gluonj.runtime.Closure"));
            ccClosure.addConstructor(CtNewConstructor
                    .defaultConstructor(ccClosure));

            CtMethod cmProceed = CtNewMethod
                    .abstractMethod(pool.get("java.lang.Object"), "proceed",
                            new CtClass[] { pool.get("java.lang.Object"),
                                    pool.get("java.lang.Object[]") },
                            new CtClass[0], ccClosure);

            cmProceed.setModifiers(Modifier.PUBLIC);
            String body = "{" + proceedBody.replaceAll("\\}", "}\\\n") + "}";

            cmProceed.setBody(body);
            ccClosure.addMethod(cmProceed);

            return ccClosure;
        } catch (CannotCompileException e) {
            throw new InternalError(e);
        } catch (NotFoundException e) {
            throw new InternalError(e);
        }

    }

    static public void print(CtClass cc) {
        System.out.println("cc=" + cc);
        System.out.println("declared");
        CtMethod[] ms = cc.getDeclaredMethods();
        print(ms);
        System.out.println("notdeclared");
        ms = cc.getMethods();
        print(ms);
    }

    static public void print(CtMethod[] ms) {
        for (int i = 0; i < ms.length; i++) {
            System.out.print(Modifier.toString(ms[i].getModifiers()) + " ");
            System.out.print(ms[i].getName() + " (");
            try {
                for (int j = 0; j < ms[i].getParameterTypes().length; j++) {
                    System.out.print(ms[i].getParameterTypes()[j].getName()
                            + ",");
                }
            } catch (NotFoundException e) {
                e.printStackTrace();
            }
            System.out.println(")");
        }

    }

    private String makeProceedBody(CtField field, boolean isSetter)
            throws NotFoundException {

        StringBuffer body = new StringBuffer();
        if (isSetter) {
            body
                    .append(makeVariableDeclCode(new CtClass[] { field
                            .getType() }));
            body.append(makeTargetTypeCheckCode(field.getDeclaringClass()));
            body.append(makeArgsSizeCheckCode(1));
            body
                    .append(makeArgsTypeCheckCode(new CtClass[] { field
                            .getType() }));
            body.append(makeSetCode(field));
        } else {
            body.append(makeTargetTypeCheckCode(field.getDeclaringClass()));
            body.append(makeArgsSizeCheckCode(0));
            body.append(makeGetCode(field));
        }
        return body.toString();
    }

    /*
     * if field type is primitivie, returns:
     * 
     * "return new <wrapper-class> ((( <declaring-type>) $1). <field-name>)";
     * 
     * if field type is object, returns:
     * 
     * "return (( <declaring-type>) $1). <field-name>";
     */
    private String makeGetCode(CtField field) throws NotFoundException {

        StringBuffer body = new StringBuffer();
        CtClass ccType = field.getType();
        String declaringClassName = field.getDeclaringClass().getName();
        
        if (ccType.isPrimitive()) {
            body.append("return new " + getWrapperName(ccType) + "(");
            body.append("  (("  + declaringClassName + ") "  + TARGET_NAME + ").");
            body.append("    " + field.getName());
            body.append(");");
        } else {
            body.append("return");
            body.append("  (("  + declaringClassName + ") "  + TARGET_NAME + ").");
            body.append("    " + field.getName());
            body.append(";");
        }
        return body.toString();
    }

    /*
     * returns:
     *  "( <declaring-type>) $1). <field-name> = args0; return null;"
     */
    private String makeSetCode(CtField field) throws NotFoundException {
        
        StringBuffer body = new StringBuffer();
        CtClass ccType = field.getType();
        body.append("((" + field.getDeclaringClass().getName() + ") " + TARGET_NAME + ").");
        body.append("   " +field.getName() + " = " + METHOD_ARGUMENT + "0;");
        body.append("return null;");
        return body.toString();
   }

    private String makeProceedBody(CtMethod callee) throws NotFoundException {

        StringBuffer body = new StringBuffer();
        body.append(makeVariableDeclCode(callee.getParameterTypes()));
        body.append(makeTargetTypeCheckCode(callee.getDeclaringClass()));
        body.append(makeArgsSizeCheckCode(callee.getParameterTypes().length));
        body.append(makeArgsTypeCheckCode(callee.getParameterTypes()));
        body.append(makeCallCode(callee));
        return body.toString();
    }

    private final String METHOD_ARGUMENT = "arg";

    private static final String TARGET_NAME = "$1"; // 1st parameter : Object

    private static final String ARGS_NAME = "$2";// 2nd parameter : Object[]

    /*
     * variable declaration. this method returns:
     * 
     * <Args-Type0> arg0; <Args-Type1> arg1; ...
     *  
     */
    private String makeVariableDeclCode(CtClass[] ccParams) {

        StringBuffer body = new StringBuffer();
        for (int i = 0; i < ccParams.length; i++) {
            String paramType = ccParams[i].getName();
            body.append(paramType + " " + METHOD_ARGUMENT + i + ";\n");
        }
        return body.toString();
    }

    private String makeArgsTypeCheckCode(CtClass[] ccParams) {

        StringBuffer body = new StringBuffer();
        /* check arguments type */
        for (int i = 0; i < ccParams.length; i++) {
            String arg = ARGS_NAME + "[" + i + "]";
            String new_arg = METHOD_ARGUMENT + i;
            CtClass ccParam = ccParams[i];

            String typeCheckCondition;
            if (ccParam.isPrimitive()) {
                typeCheckCondition = arg + " instanceof "
                        + getWrapperName(ccParam);
            } else {
                typeCheckCondition = arg + " instanceof " + ccParam.getName();
            }
            String substitution;
            if (ccParam.isPrimitive()) {
                substitution = new_arg + " = ((" + getWrapperName(ccParam)
                        + ")" + arg + ")." + ccParam.getName() + "Value();\n";
            } else {
                substitution = new_arg + " = (" + ccParam.getName() + ")" + arg
                        + ";";
            }

            body.append("if(" + typeCheckCondition + ") {");
            body.append(substitution);
            body.append("} else {");
            body.append("throw new gluonj.runtime.UnmatchParameterException(");
            body.append("\"" + ccParam.getName() + "\", " + arg
                    + ".getClass().getName());");
            body.append("}");
        }
        return body.toString();
    }

    private String makeArgsSizeCheckCode(int paramNum) {

        StringBuffer body = new StringBuffer();

        /* check number of arguments */
        body.append("if(!(((Object[])" + ARGS_NAME + ").length == " + paramNum
                + ")) {");
        body
                .append("throw new gluonj.runtime.UnmatchParameterException(\"parameter num must be "
                        + paramNum + ".\");");
        body.append("}");
        return body.toString();
    }

    /*
     * returns:
     * 
     * if(!$1 instanceof <this-Class>)
     * 
     * throw new UnmatchParamterException(" <this-Class>",
     * this.getClass().getName());
     *  
     */
    private String makeTargetTypeCheckCode(CtClass ccDeclaring) {
        StringBuffer body = new StringBuffer();
        String calleeDeclClass = ccDeclaring.getName();

        body.append("if(!(" + TARGET_NAME + " instanceof " + calleeDeclClass
                + ")) {");
        body.append("throw new gluonj.runtime.UnmatchParameterException(");
        body.append("\"" + calleeDeclClass + "\", " + TARGET_NAME
                + ".getClass().getName());");
        body.append("}");
        return body.toString();
    }

    private String makeCallCode(CtMethod callee) throws NotFoundException {

        String calleeClass = callee.getDeclaringClass().getName();
        int paramNum = callee.getParameterTypes().length;
        String returnType = callee.getReturnType().getName();
        // argsSet = (a0, a1, ...)
        String argsSet = argsSet(paramNum);

        StringBuffer body = new StringBuffer();
        CtClass ccRetType = callee.getReturnType();

        // callCode = ((<callee-class>) $0).<method>(a1, a2, ...)
        String callCode = "((" + calleeClass + ")" + TARGET_NAME + ")."
                + callee.getName() + argsSet;

        if (ccRetType.equals(CtClass.voidType)) {
            body.append(callCode + ";");
            body.append("return null;");
        } else if (ccRetType.isPrimitive()) {
            String wrapperName = getWrapperName(ccRetType);
            body.append("return new " + wrapperName + "(");
            body.append(callCode);
            body.append(");");
        } else {
            body.append("return " + callCode + ";");
        }
        return body.toString();
    }

    /**
     * @param num
     * @return String object such as "(a0, a1, a2)", if num is 2.
     */
    private String argsSet(int num) {
        StringBuffer argsSet = new StringBuffer("(");
        if (num > 0) {
            argsSet.append(METHOD_ARGUMENT + "0");
        }
        for (int i = 1; i < num; i++) {
            argsSet.append("," + METHOD_ARGUMENT + i);
        }
        argsSet.append(")");
        return argsSet.toString();
    }

    private String makeClosureName(CtClass target, CtMethod callee) {

        int h = Math.abs(callee.hashCode());
        return "Closure_" + callee.getName() + h;
    }

    private String makeClosureName(CtClass target, CtField field,
            boolean isSetter) {
        int h = Math.abs(field.hashCode());
        String accessType = isSetter ? "set" : "get";
        return "Closure_" + field.getName() + "_" + accessType + "_" + h;

    }

    private String getWrapperName(CtClass cc) {

        if (wrapperTable.containsKey(cc)) {
            String wrapperName = (String) wrapperTable.get(cc);
            return wrapperName;
        } else {
            return null;
        }
    }

    private Map wrapperTable = new HashMap();
    {
        wrapperTable.put(CtClass.booleanType, "java.lang.Boolean");
        wrapperTable.put(CtClass.byteType, "java.lang.Byte");
        wrapperTable.put(CtClass.charType, "java.lang.Character");
        wrapperTable.put(CtClass.doubleType, "java.lang.Double");
        wrapperTable.put(CtClass.floatType, "java.lang.Float");
        wrapperTable.put(CtClass.intType, "java.lang.Integer");
        wrapperTable.put(CtClass.longType, "java.lang.Long");
        wrapperTable.put(CtClass.shortType, "java.lang.Short");
        wrapperTable.put(CtClass.voidType, "java.lang.Void");
    }
}