/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.weaver.gen;

import gluonj.ast.Aspect;
import gluonj.ast.PointcutDecl;
import gluonj.ast.java.ASTFormalParameter;
import gluonj.ast.pointcut.ASTArgs;
import gluonj.ast.pointcut.ASTPCReference;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.ast.pointcut.ASTPointcutPrimitive;
import gluonj.ast.pointcut.ASTTarget;
import gluonj.ast.pointcut.ASTThis;
import gluonj.ast.pointcut.visitor.PcPrimitiveTranslator;
import gluonj.ast.pointcut.visitor.Traverser;
import gluonj.weaver.CannotCompileException;
import gluonj.weaver.JoinPointConstants;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class NamedPCProcessor extends AbstractProcessor {

    private List pcdeclList;

    public NamedPCProcessor() {
        pcdeclList = new LinkedList();
    }

    public Aspect[] process(Aspect[] aspects) throws CannotCompileException {

        Aspect[] new_aspects = new Aspect[aspects.length];
        for (int i = 0; i < aspects.length; i++) {
            new_aspects[i] = process(aspects[i]);
        }
        return new_aspects;
    }

    protected Aspect process(Aspect aspect) throws CannotCompileException {
        PointcutDecl[] decls = aspect.getPointcutDecl();
        pcdeclList.addAll(Arrays.asList(decls));
        return super.process(aspect);
    }

    protected ASTPointcutExpr process(ASTPointcutExpr expr)
            throws CannotCompileException {

        ReferenceRemover v = new ReferenceRemover(pcdeclList);
        try {
            Traverser.traverseAndOrNotPrimitive(expr, v);
        } catch (RuntimeException e) {
            if (e.getCause() instanceof CannotCompileException) {
                throw (CannotCompileException) e.getCause();
            } else {
                throw e;
            }
        }
        return v.getNewPC();
    }

    private static class ReferenceRemover extends PcPrimitiveTranslator {

        List pcdeclList;

        ParameterNameTable nameTable;

        public ReferenceRemover(List l) {
            pcdeclList = l;
            nameTable = new ParameterNameTable();
        }

        ReferenceRemover(List l, ParameterNameTable table) {
            pcdeclList = l;
            nameTable = table;
        }

        public ASTPointcutExpr changePcPrimitive(ASTPointcutPrimitive e) {

            if (e.getType() == JoinPointConstants.THIS) {
                ASTThis expr = (ASTThis) e;
                return changeThis(expr);
            } else if (e.getType() == JoinPointConstants.ARGS) {
                ASTArgs expr = (ASTArgs) e;
                return changeArgs(expr);
            } else if (e.getType() == JoinPointConstants.TARGET) {
                ASTTarget expr = (ASTTarget) e;
                return changeTarget(expr);
            } else if (e.getType() == JoinPointConstants.REFERENCE) {
                ASTPCReference expr = (ASTPCReference) e;
                return changeReference(expr);
            } else {
                return e;
            }
        }

        private ASTPointcutExpr changeReference(ASTPCReference expr) {
            PointcutDecl referredPc = null;

            for (Iterator iter = pcdeclList.iterator(); iter.hasNext();) {
                PointcutDecl decl = (PointcutDecl) iter.next();
                if (decl.getName().equals(expr.getName())) {
                    referredPc = decl;
                    break;
                }
            }
            if (referredPc == null) {
                throw new RuntimeException(new CannotCompileException(
                        "There is no pointcut declaration which name is \""
                                + expr.getName() + "\"."));
            }

            String[] argNames = expr.getArguments();
            String[] paramNames = new String[referredPc.getParametersList()
                    .size()];

            int i = 0;
            for (Iterator iter = referredPc.getParametersList().iterator(); iter
                    .hasNext();) {
                ASTFormalParameter param = (ASTFormalParameter) iter.next();
                paramNames[i] = param.getName();
                i++;
            }
            nameTable.addNewParameterRelations(paramNames, argNames);
            ReferenceRemover new_visitor = new ReferenceRemover(pcdeclList,
                    nameTable);
            Traverser.traverseAndOrNotPrimitive(referredPc.getPointcut(),
                    new_visitor);

            ASTPointcutExpr inlinedExpr = new_visitor.getNewPC();
            nameTable.removePreviousParameterRelations();

            return inlinedExpr;
        }

        private ASTPointcutExpr changeTarget(ASTTarget expr) {
            if (expr.isBinder()) {
                String new_varname = nameTable.find(expr.getVarName());
                return new ASTTarget(new_varname, expr.getTypeName());
            } else {
                return expr;
            }
        }

        private ASTPointcutExpr changeArgs(ASTArgs expr) {
            if (expr.isBinder()) {
                String[] new_name = new String[expr.getVarName().length];
                for (int i = 0; i < new_name.length; i++) {
                    new_name[i] = nameTable.find(expr.getVarName()[i]);
                }
                return new ASTArgs(new_name, expr.getTypeName());
            } else {
                return expr;
            }
        }

        private ASTPointcutExpr changeThis(ASTThis expr) {
            if (expr.isBinder()) {
                String new_varname = nameTable.find(expr.getVarName());
                return new ASTThis(new_varname, expr.getTypeName());
            } else {
                return expr;
            }
        }
    }

    private static class ParameterNameTable {

        List mapStack = new LinkedList(); // List<Map<String, String>>

        String find(String name) {
            for (Iterator iter = mapStack.iterator(); iter.hasNext();) {
                Map map = (Map) iter.next();
                if (map.containsKey(name)) {
                    name = (String) map.get(name);
                } else {
                    throw new RuntimeException(new CannotCompileException(
                            "Variable " + name + " is not declared."));
                }
            }
            return name;
        }

        void addNewParameterRelations(String[] paramNames, String[] argNames) {
            if (paramNames.length != argNames.length) {
                throw new RuntimeException();
            }
            Map newMap = new HashMap();
            for (int i = 0; i < paramNames.length; i++) {
                newMap.put(paramNames[i], argNames[i]);
            }
            mapStack.add(0, newMap);
        }

        void removePreviousParameterRelations() {
            mapStack.remove(0);
        }

    }
}