/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.weaver.gen;

import gluonj.ast.joinpoint.JoinPoint;
import gluonj.ast.pcs.IfResidue;
import gluonj.ast.pcs.PcsBehavior;
import gluonj.ast.pcs.PcsCall;
import gluonj.ast.pcs.PcsModifier;
import gluonj.ast.pcs.PcsOr;
import gluonj.ast.pcs.PcsParameter;
import gluonj.ast.pcs.PcsParameterDotDot;
import gluonj.ast.pcs.PcsParameterType;
import gluonj.ast.pcs.PcsType;
import gluonj.ast.pcs.PointcutResidue;
import gluonj.ast.pcs.PointcutShadow;
import gluonj.ast.pointcut.ASTArgs;
import gluonj.ast.pointcut.ASTNot;
import gluonj.ast.pointcut.ASTPCShadow;
import gluonj.ast.pointcut.ASTPointcutExpr;
import gluonj.ast.pointcut.ASTPointcutPrimitive;
import gluonj.ast.pointcut.ASTTarget;
import gluonj.ast.pointcut.ASTThis;
import gluonj.ast.pointcut.ASTUnaryExpr;
import gluonj.ast.pointcut.visitor.DNFTranslator;
import gluonj.ast.pointcut.visitor.Traverser;
import gluonj.weaver.JoinPointConstants;
import gluonj.weaver.UnSupportedException;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/*
 * pcd of this, args, target Aresidue pcd (if pcd) ֕ϊ
 * 
 * (exec && within && this) || (call && target && args) ->
 * 
 * (exec && within && residue) || (call && residue)
 */
public class ResidueProcessor extends AbstractProcessor {

    protected ASTPointcutExpr process(ASTPointcutExpr expr) {

        DNFTranslator t = new ResidueTranslator();
        Traverser.traverseDNF(expr, t);
        return t.getDNF();
    }

    private static class ResidueTranslator extends DNFTranslator {

        public List[] translateCNF(List exprs) {
            List newCNF = new LinkedList();

            for (Iterator iter = exprs.iterator(); iter.hasNext();) {

                ASTUnaryExpr e = (ASTUnaryExpr) iter.next();
                int type = e.getType();
                ASTPointcutExpr newExpr;

                if (type == JoinPointConstants.NOT) {
                    newExpr = changeNot((ASTNot) e);
                } else {
                    newExpr = changePcPrimitive((ASTPointcutPrimitive) e, false);
                }
                newCNF.add(newExpr);
            }
            return new List[] { newCNF };
        }

        public ASTPointcutExpr changeNot(ASTNot not) {
            ASTPointcutPrimitive e = (ASTPointcutPrimitive) not.getChild();
            return changePcPrimitive(e, true);
        }

        public ASTPointcutExpr changePcPrimitive(ASTPointcutPrimitive a,
                boolean isNot) {

            int type = a.getType();
            if (type == JoinPointConstants.ARGS) {
                return changeArgs((ASTArgs) a, isNot);
            } else if (type == JoinPointConstants.THIS) {
                return changeThis((ASTThis) a, isNot);
            } else if (type == JoinPointConstants.TARGET) {
                return changeTarget((ASTTarget) a, isNot);
            } else if (type == JoinPointConstants.SHADOW) {
                ASTPCShadow pc = ((ASTPCShadow) a);
                if (pc.getShadow().equals(PointcutShadow.TRUE)
                        && pc.getResidue() instanceof IfResidue && isNot) {
                    String cond = ((IfResidue) pc.getResidue())
                            .getDynamicCondition();
                    return new ASTPCShadow(PointcutShadow.TRUE, new IfResidue(
                            "!(" + cond + ")"));
                } else {
                    return pc;
                }
            } else if (isNot) {
                return new ASTNot(a);
            } else {
                return a;
            }
        }

        public ASTPointcutExpr changeThis(ASTThis expr, boolean isNot) {
            if (isNot && expr.isBinder()) {
                throw new UnSupportedException("negation doesn't allow binding");
            }
            PointcutResidue r = makeThisResidue(expr, isNot);
            return new ASTPCShadow(PointcutShadow.TRUE, r);

        }

        private PointcutResidue makeThisResidue(ASTThis expr, boolean isNot) {
            final String stmt;
            final String cond;
            if (expr.isBinder()) {
                stmt = expr.getVarName() + " = " + "this" + ";\n";
            } else {
                stmt = "";
            }
            if (isNot) {
                cond = "!(this" + " instanceof " + expr.getTypeName() + ")";
            } else {
                cond = "this" + " instanceof " + expr.getTypeName();
            }
            return new PointcutResidue() {
                public String getDynamicCondition(JoinPoint jp) {
                    return cond;
                }

                public String getStatement(JoinPoint jp) {
                    return stmt;
                }
            };
        }

        public ASTPointcutExpr changeTarget(ASTTarget expr, boolean isNot) {
            if (isNot && expr.isBinder()) {
                throw new UnSupportedException("negation doesn't allow binding");
            }
            PointcutResidue r = makeTargetResidue(expr, isNot);
            return new ASTPCShadow(PointcutShadow.TRUE, r);
        }

        private PointcutResidue makeTargetResidue(ASTTarget expr, boolean isNot) {
            return new TargetResidue(expr, isNot);
        }

        private class TargetResidue extends PointcutResidue {

            boolean isNot;

            ASTTarget expr;

            TargetResidue(ASTTarget e, boolean b) {
                isNot = b;
                expr = e;
            }

            public String getDynamicCondition(JoinPoint jp) {
                if (jp.getJPType() == JoinPoint.METHOD_CALL) {
                    return makeCondition("$0");
                } else {
                    return makeCondition("this");
                }
            }

            public String getStatement(JoinPoint jp) {
                if (jp.getJPType() == JoinPoint.METHOD_CALL) {
                    return makeStatement("$0");
                } else {
                    return makeStatement("this");
                }
            }

            private String makeCondition(String target) {
                String cond = target + " instanceof " + expr.getTypeName();
                if (isNot) {
                    return "!(" + cond + ")";
                } else {
                    return cond;
                }
            }

            private String makeStatement(String target) {
                if (expr.isBinder()) {
                    return expr.getVarName() + " = " + target + ";\n";
                } else {
                    return "";
                }
            }
        }

        public ASTPointcutExpr changeArgs(ASTArgs expr, boolean isNot) {

            if (isNot && expr.isBinder()) {
                throw new UnSupportedException("negation doesn't allow binding");
            }
            PointcutShadow shadow = makeArgsStaticCondition(expr, isNot);
            PointcutResidue residue = makeArgsResidue(expr, isNot);
            return new ASTPCShadow(shadow, residue);
        }

        private PointcutResidue makeArgsResidue(ASTArgs expr, boolean isNot) {
            final String cond;
            final String stmt;
            if (isNot) {
                cond = "!(" + makeArgsDynamicCondition(expr) + ")";
            } else {
                cond = makeArgsDynamicCondition(expr);
            }
            if (expr.isBinder()) {
                stmt = makeArgsStatement(expr);
            } else {
                stmt = "";
            }
            PointcutResidue r = new PointcutResidue() {
                public String getDynamicCondition(JoinPoint jp) {
                    return cond;
                }

                public String getStatement(JoinPoint jp) {
                    return stmt;
                }
            };
            return r;
        }

        private String makeArgsDynamicCondition(ASTArgs expr) {
            StringBuffer cond = new StringBuffer("true");

            int idx = 0;
            for (idx = 0; idx < expr.getTypeName().length; idx++) {

                String name = expr.getTypeName()[idx];
                if (name.equals("..")) {
                    break;
                }
                String arg = "$args[" + idx + "]";
                cond.append("&& (" + arg + " instanceof " + name + ")");
            }

            for (; idx < expr.getTypeName().length; idx++) {

                String name = expr.getTypeName()[idx];
                String arg = "$args[ $args.length - "
                        + (expr.getTypeName().length - idx) + "])";
                cond.append("&& (" + arg + " instanceof " + name + ")");
            }
            return cond.toString();
        }

        private String makeArgsStatement(ASTArgs expr) {

            StringBuffer statement = new StringBuffer();
            int idx = 0;
            for (idx = 0; idx < expr.getTypeName().length; idx++) {

                String name = expr.getTypeName()[idx];
                if (name.equals("..")) {
                    break;
                }
                String arg = "$args[" + idx + "]";
                String var = expr.getVarName()[idx];
                String castedArg = "((" + name + ")" + arg + ")";
                statement.append(var + " = " + castedArg + ";");
            }

            for (; idx < expr.getTypeName().length; idx++) {

                String name = expr.getTypeName()[idx];
                String arg = "$args[ $args.length - "
                        + (expr.getTypeName().length - idx) + "])";
                String var = expr.getVarName()[idx];
                String castedArg = "((" + name + ")" + arg + ")";
                statement.append(var + " = " + arg + ";");
            }

            return "{" + statement.toString() + "\n}";
        }

        private PointcutShadow makeArgsStaticCondition(ASTArgs expr,
                boolean isNot) {
            // make static condition
            PcsParameter[] paramPat = new PcsParameter[expr.getTypeName().length];
            for (int i = 0; i < paramPat.length; i++) {
                if (expr.getTypeName()[i].equals("..")) {
                    paramPat[i] = new PcsParameterDotDot();
                } else {
                    paramPat[i] = new PcsParameterType(PcsType.ANY_TYPE);
                }
            }
            PcsBehavior argsPat = new PcsBehavior(new PcsModifier(),
                    PcsType.ANY_TYPE, paramPat, new PcsType[0]);

            PointcutShadow shadow = new PcsOr(argsPat, new PcsCall(argsPat));
            return shadow;
        }
    }
}