/*
 * GluonJ 2005, an aspect-oriented system for Java.
 * Copyright (C) 2006 Rei Ishikawa.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 */
package gluonj.ast.pcs;

import gluonj.ast.joinpoint.JPFactory;
import gluonj.ast.pattern.ArrayTypeNamePattern;
import gluonj.ast.pattern.BasicTypeNamePattern;
import gluonj.ast.pattern.DotTypeNamePattern;
import gluonj.ast.pattern.IdPattern;
import gluonj.ast.pattern.IdentifierTypeNamePattern;
import gluonj.ast.pattern.PrimitiveTypeNamePattern;
import gluonj.ast.pattern.TypeNamePattern;

import java.lang.reflect.Modifier;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.FieldAccess;
import javassist.expr.MethodCall;
import test.GluonJTestCase;


public class TestMatchingOfPcShadow extends GluonJTestCase {


	/*
	 * target: java.lang.String pattern: java.*.String
	 */

	public final void testClass1() throws NotFoundException {

		ClassPool pool = getPool(); 
		CtClass string = pool.get("java.lang.String");
		PcsType t = makeStringTypePat();

		
		

		assertTrue(t.match(string));
	}

	/*
	 * java.*.String
	 */
	static PcsType makeStringTypePat() {
		BasicTypeNamePattern tn = new DotTypeNamePattern(new DotTypeNamePattern(
				new IdentifierTypeNamePattern(new IdPattern("java")),
				new IdentifierTypeNamePattern(new IdPattern("*"))),
				new IdentifierTypeNamePattern(new IdPattern("String")));

		PcsType t = new PcsType();
		t.name = tn;
		t.modifier = new PcsModifier();
		t.modifier.setModifiers(Modifier.PUBLIC, true);
		t.modifier.setModifiers(Modifier.ABSTRACT, false);
		t.modifier.setModifiers(Modifier.FINAL, true);
		return t;
	}

	/*
	 * target: boolean java.io.File.createNewFile() throws IOException
	 * 
	 * pattern1: * java.*.File.create*() throws java.io.IOException
	 * 
	 * pattern2: public boolean java.*.File.c*New*()
	 */
	public final void testMethod1() throws NotFoundException {

		ClassPool pool = getPool(); 
		CtClass c = pool.get("java.io.File");
		CtMethod m = c.getDeclaredMethod("createNewFile");

		/* java.io.File */
		PcsType file = makeFileTypePat();
		/* java.io.IOException */
		PcsType ioexception = makeIOExceptionPat();
		/* * */
		PcsType all = makeAllAcceptPat();
		/* method:create*() */
		PcsMethod mp = new PcsMethod();
		mp.declaredClass = file;
		mp.modifiers = new PcsModifier();
		mp.name = new IdPattern("create*");
		mp.parameter = new PcsParameter[0];
		mp.returnType = all;

		mp.throwable = new PcsType[] { ioexception };


		assertTrue(mp.match(m));

		/* pattern2: public boolean java.*.File.c*New*() */

		PcsType booleant = makeBooleanPat();

		PcsMethod mp2 = new PcsMethod();
		mp2.declaredClass = file;
		mp2.modifiers = new PcsModifier();
		mp2.modifiers.setModifiers(Modifier.PUBLIC, true);
		mp2.name = new IdPattern("c*New*");
		mp2.parameter = new PcsParameter[0];
		mp2.returnType = booleant;
		mp2.throwable = new PcsType[0];

		assertTrue(mp2.match(m));
	}

	/*
	 * boolean
	 */
	static PcsType makeBooleanPat() {
		BasicTypeNamePattern booleantn = new PrimitiveTypeNamePattern(PrimitiveTypeNamePattern.BOOLEAN);
		PcsType booleant = new PcsType();
		booleant.modifier = new PcsModifier();
		booleant.name = booleantn;
		return booleant;
	}

	/*
	 * *
	 */
	static PcsType makeAllAcceptPat() {
		PcsType all = new PcsType();
		all.modifier = new PcsModifier();
		all.name = new IdentifierTypeNamePattern(
				new IdPattern("*"));
		return all;
	}

	/*
	 * java.*.File
	 */
	static PcsType makeFileTypePat() {
		BasicTypeNamePattern tn = new DotTypeNamePattern(new DotTypeNamePattern(
				new IdentifierTypeNamePattern(new IdPattern("java")),
				new IdentifierTypeNamePattern(new IdPattern("*"))),
				new IdentifierTypeNamePattern(new IdPattern("File")));

		PcsType file = new PcsType();
		file.name = tn;
		file.modifier = new PcsModifier();
		return file;
	}

	/*
	 * java.io.IOException
	 */
	static PcsType makeIOExceptionPat() {
		BasicTypeNamePattern ioex = new DotTypeNamePattern(new DotTypeNamePattern(
				new IdentifierTypeNamePattern(new IdPattern("java")),
				new IdentifierTypeNamePattern(new IdPattern("io"))),
				new IdentifierTypeNamePattern(new IdPattern("IOException")));

		PcsType ioexception = new PcsType();
		ioexception.name = ioex;
		ioexception.modifier = new PcsModifier();

		return ioexception;
	}

	/*
	 * target: public java.lang.String(byte[], int, int)
	 * 
	 * 
	 * pattern: public java.lang.String(byte[], int, int)
	 * 
	 * target: public java.lang.String(byte[], int, int)
	 * 
	 * 
	 * pattern: public java.lang.String(int, int, byte[])
	 */

	public final void testConstructor() throws NotFoundException {

		ClassPool pool = getPool(); 
		CtConstructor cc = pool.get("java.lang.String").getDeclaredConstructor(
				new CtClass[] { pool.get("byte[]"), pool.get("int"),
						pool.get("int"), });
		PcsConstructor cp1 = new PcsConstructor();

		TypeNamePattern stringtnp = 
				new DotTypeNamePattern(new DotTypeNamePattern(
						new IdentifierTypeNamePattern(new IdPattern("java")),
						new IdentifierTypeNamePattern(new IdPattern("lang"))),
						new IdentifierTypeNamePattern(new IdPattern("String")));

		TypeNamePattern bytearray = 
				new ArrayTypeNamePattern(new PrimitiveTypeNamePattern(PrimitiveTypeNamePattern.BYTE));
		TypeNamePattern intt = 
				new PrimitiveTypeNamePattern(PrimitiveTypeNamePattern.INT);

		PcsType param1 = new PcsType();
		param1.modifier = new PcsModifier();
		param1.name = bytearray;
		PcsType param2 = new PcsType();
		param2.modifier = new PcsModifier();
		param2.name = intt;

		cp1.declaredClass = new PcsType();
		cp1.declaredClass.modifier = new PcsModifier();
		cp1.declaredClass.name = stringtnp;

		cp1.modifiers = new PcsModifier();
		cp1.modifiers.setModifiers(Modifier.PUBLIC, true);

		cp1.parameter = new PcsParameter[] {
				new PcsParameterType(param1),
				new PcsParameterType(param2),
				new PcsParameterType(param2), };

		cp1.throwable = new PcsType[0];

		assertTrue(cp1.match(cc));

		cp1.parameter = new PcsParameter[] {
				new PcsParameterType(param2),
				new PcsParameterType(param2),
				new PcsParameterType(param1), };

		
		assertFalse(cp1.match(cc));
	}

	/*
	 * IN public void foo() IN public gluonj.ast.pattern.A
	 * 
	 * pattern: set(java.*.* gluonj.*.*.A.str) && within(public * *.foo*())
	 * pattern: get(java.*.*Stream java.lang.System.out) && within(public *
	 * *.foo*())
	 * 
	 *  pattern: call(void java.*.*Stream.println(String)) &&
	 * within(public * *.foo*())
	 */
	public final void testFieldAccess() throws NotFoundException,
			CannotCompileException {

		ClassPool pool = getPool(); 
		CtMethod a = pool.get("gluonj.ast.pattern.A").getDeclaredMethod("foo");
		final PcsSet set = new PcsSet();
		final StringBuffer result = new StringBuffer();

		set.field = new PcsField();
		set.field.declaredClass = makeAPat();
		set.field.modifiers = new PcsModifier();
		set.field.name = new IdPattern("str");
		set.field.type = makeStringPat2();

		set.where = new PcsMethod();
		((PcsMethod) set.where).returnType = makeAllAcceptPat();
		((PcsMethod) set.where).name = new IdPattern("foo*");
		set.where.declaredClass = makeAPat();
		set.where.modifiers = new PcsModifier();
		set.where.modifiers.setModifiers(Modifier.PUBLIC, true);
		set.where.parameter = new PcsParameter[0];
		set.where.throwable = new PcsType[0];

		a.instrument(new ExprEditor() {
			public void edit(FieldAccess arg0) throws CannotCompileException {

				boolean b = set.match(JPFactory.makeJP(arg0));
				if (b) {
					result.delete(0, result.length());
					result.append(true);
				}
			}
		});
		assertTrue(Boolean.valueOf(result.toString()).booleanValue());

		final PcsGet get = new PcsGet();
		get.field = new PcsField();
		get.field.declaredClass = makeSystemPat();
		get.field.modifiers = new PcsModifier();
		get.field.modifiers.setModifiers(Modifier.FINAL, true);
		get.field.name = new IdPattern("out");
		get.field.type = makeStreamPat();

		get.where = set.where;

		result.delete(0, result.length());
		a.instrument(new ExprEditor() {
			public void edit(FieldAccess arg0) throws CannotCompileException {

				boolean b = get.match(JPFactory.makeJP(arg0));
				if (b) {
					result.delete(0, result.length());
					result.append(true);
				}
			}
		});
		assertTrue(Boolean.valueOf(result.toString()).booleanValue());

		final PcsCall call = new PcsCall();
		PcsMethod callee = new PcsMethod();
		call.callee = callee;
		call.callee.declaredClass = makeStreamPat();
		call.callee.modifiers = new PcsModifier();
		callee.name = new IdPattern("print*n");
		call.callee.parameter = new PcsParameter[]{
		new PcsParameterType(makeStringTypePat())};
		call.callee.throwable = new PcsType[0];
		callee.returnType = new PcsType();
		callee.returnType.modifier = new PcsModifier();
		callee.returnType.name = 
				new PrimitiveTypeNamePattern(PrimitiveTypeNamePattern.VOID);

		call.where = get.where;

		result.delete(0, result.length());
		a.instrument(new ExprEditor() {
			public void edit(MethodCall arg0) throws CannotCompileException {

				
				boolean b = call.match(JPFactory.makeJP(arg0));
				if (b) {
					result.delete(0, result.length());
					result.append(true);
				}
			}
		});
		assertTrue(Boolean.valueOf(result.toString()).booleanValue());
	}

	/*
	 * java.*.*Stream
	 */
	static PcsType makeStreamPat() {
		BasicTypeNamePattern tn = new DotTypeNamePattern(new DotTypeNamePattern(
				new IdentifierTypeNamePattern(new IdPattern("java")),
				new IdentifierTypeNamePattern(new IdPattern("*"))),
				new IdentifierTypeNamePattern(new IdPattern("*Stream")));

		PcsType t = new PcsType();
		t.name = tn;
		t.modifier = new PcsModifier();
		return t;
	}

	/*
	 * java.lang.System
	 */
	static PcsType makeSystemPat() {
		BasicTypeNamePattern tn = new DotTypeNamePattern(new DotTypeNamePattern(
				new IdentifierTypeNamePattern(new IdPattern("java")),
				new IdentifierTypeNamePattern(new IdPattern("lang"))),
				new IdentifierTypeNamePattern(new IdPattern("System")));

		PcsType t = new PcsType();
		t.name = tn;
		t.modifier = new PcsModifier();
		return t;
	}

	/*
	 * java.*.*
	 */
	static PcsType makeStringPat2() {
		BasicTypeNamePattern tn = new DotTypeNamePattern(new DotTypeNamePattern(
				new IdentifierTypeNamePattern(new IdPattern("java")),
				new IdentifierTypeNamePattern(new IdPattern("*"))),
				new IdentifierTypeNamePattern(new IdPattern("*")));

		PcsType t = new PcsType();
		t.name = tn;
		t.modifier = new PcsModifier();
		return t;
	}

	/*
	 * gluonj.*.*.A
	 */
	static PcsType makeAPat() {
		BasicTypeNamePattern tn = new DotTypeNamePattern(new DotTypeNamePattern(
				new DotTypeNamePattern(new IdentifierTypeNamePattern(
						new IdPattern("gluonj")),
						new IdentifierTypeNamePattern(new IdPattern("*"))),
				new IdentifierTypeNamePattern(new IdPattern("*"))),
				new IdentifierTypeNamePattern(new IdPattern("A")));

		PcsType t = new PcsType();
		t.name = tn;
		t.modifier = new PcsModifier();
		return t;
	}

}